#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer macro
manager"""

__all__ = ["MacroManager", "MacroExecutor", "is_macro"]

__docformat__ = 'restructuredtext'

import re
import os
import sys
import copy
import inspect
import logging
import functools
import traceback
import threading

from lxml import etree

import time

from PyTango import DevFailed

from collections import OrderedDict

from taurus.core.util.log import Logger
from taurus.core.util.codecs import CodecFactory

from sardana.sardanadefs import ElementType
from sardana.sardanamodulemanager import ModuleManager
from sardana.sardanaexception import format_exception_only_str
from sardana.sardanautils import is_pure_str, is_non_str_seq, recur_map

from sardana.macroserver.msmanager import MacroServerManager
from sardana.macroserver.msmetamacro import MACRO_TEMPLATE, MacroLibrary, \
    MacroClass, MacroFunction
from sardana.macroserver.msparameter import ParamDecoder, FlatParamDecoder, \
    WrongParam
from sardana.macroserver.macro import Macro, MacroFunc, ExecMacroHook, \
    Hookable
from sardana.macroserver.msexception import UnknownMacroLibrary, \
    LibraryError, UnknownMacro, MissingEnv, AbortException, StopException, \
    ReleaseException, MacroServerException, UnknownEnv
from sardana.util.parser import ParamParser
from sardana.util.thread import raise_in_thread

# These classes are imported from the "client" part of sardana, if finally
# both the client and the server side needs them, place them in some
# common location
from sardana.taurus.core.tango.sardana.macro import createMacroNode


_BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def islambda(f):
    """inspect doesn't come with islambda so I create one :-P"""
    return inspect.isfunction(f) and \
        f.__name__ == (lambda: True).__name__


def _is_macro(macro, abs_file=None, logger=None):
    """Helper function to determine if a certain python object is a valid
    macro"""
    if inspect.isclass(macro):
        if not issubclass(macro, Macro):
            return False
        # if it is a class defined in some other module forget it to
        # avoid replicating the same macro in different macro files

        # use normcase to treat case insensitivity of paths on
        # certain platforms e.g. Windows
        if os.path.normcase(inspect.getabsfile(macro)) !=\
           os.path.normcase(abs_file):
            return False

    elif callable(macro) and not islambda(macro):
        # if it is a function defined in some other module forget it to
        # avoid replicating the same macro in different macro files

        # use normcase to treat case insensitivity of paths on
        # certain platforms e.g. Windows
        if os.path.normcase(inspect.getabsfile(macro)) !=\
           os.path.normcase(abs_file):
            return False

        if not hasattr(macro, 'macro_data'):
            return False

        args, varargs, keywords, *_ = inspect.getfullargspec(macro)
        if len(args) == 0:
            if logger:
                logger.debug("Could not add macro %s: Needs at least one "
                             "parameter (usually called 'self')",
                             macro.__name__)
            return False
        if keywords is not None:
            if logger:
                logger.debug("Could not add macro %s: Unsupported keyword "
                             "parameters '%s'", macro.__name__, keywords)
            return False
        if varargs and len(args) > 1:
            if logger:
                logger.debug("Could not add macro %s: Unsupported giving "
                             "named parameters '%s' and varargs '%s'",
                             macro.__name__, args, varargs)
            return False
    else:
        return False
    return True


def is_macro(macro, abs_file=None, logger=None):
    try:
        return _is_macro(macro, abs_file=abs_file, logger=logger)
    except Exception:
        return False


def is_flat_list(obj):
    """Check if a given object is a flat list."""
    if not isinstance(obj, list):
        return False
    for item in obj:
        if isinstance(item, list):
            return False
    return True


class MacroManager(MacroServerManager):

    DEFAULT_MACRO_DIRECTORIES = os.path.join(_BASE_DIR, 'macros'),

    def __init__(self, macro_server, macro_path=None):
        MacroServerManager.__init__(self, macro_server)
        if macro_path is not None:
            self.setMacroPath(macro_path)

    def reInit(self):
        if self.is_initialized():
            return

        # dict<str, MacroLibrary>
        # key   - module name (without path and without extension)
        # value - MacroLibrary object representing the module
        self._modules = {}

        # dict<str, <MacroClass>
        # key   - macro name
        # value - MacroClass object representing the macro
        self._macro_dict = {}

        # list<str>
        # elements are absolute paths
        self._macro_path = []

        # list<str>
        # overwritten macros (macros with the same name defined in
        # different modules)
        self._overwritten_macros = []

        # dict<Door, <MacroExecutor>
        # key   - door
        # value - MacroExecutor object for the door
        self._macro_executors = {}

        MacroServerManager.reInit(self)

    def cleanUp(self):
        if self.is_cleaned():
            return

        # if self._modules:
        #    ModuleManager().unloadModules(self._modules.keys())

        self._macro_path = None
        self._macro_dict = None
        self._modules = None
        self._overwritten_macros = None

        MacroServerManager.cleanUp(self)

    def setMacroPath(self, macro_path):
        """Registers a new list of macro directories in this manager.
        Warning: as a consequence all the macro modules will be reloaded.
        This means that if any reference to an old macro object was kept it will
        refer to an old module (which could possibly generate problems of type
        class A != class A)."""
        p = []
        for item in macro_path:
            p.extend(item.split(os.pathsep))

        # filter empty and commented paths
        p = [i for i in p if i and not i.startswith("#")]

        # add basic macro directories
        for macro_dir in self.DEFAULT_MACRO_DIRECTORIES:
            if macro_dir not in p:
                p.append(macro_dir)

        self._macro_path = p

        macro_file_names = self._findMacroLibNames()
        for mod_name, file_name in macro_file_names.items():
            dir_name = os.path.dirname(file_name)
            path = [dir_name]
            try:
                self.reloadMacroLib(mod_name, path)
            except:
                pass

    def getMacroPath(self):
        return self._macro_path

    def _findMacroLibName(self, lib_name, path=None):
        path = path or self.getMacroPath()
        f_name = lib_name
        if not f_name.endswith('.py'):
            f_name += '.py'
        for p in path:
            try:
                elems = os.listdir(p)
                if f_name in elems:
                    return os.path.abspath(os.path.join(p, f_name))
            except:
                self.debug("'%s' is not a valid path" % p)
        return None

    def _findMacroLibNames(self, path=None):
        path = path or self.getMacroPath()
        ret = OrderedDict()
        for p in reversed(path):
            try:
                for f in os.listdir(p):
                    name, ext = os.path.splitext(f)
                    if name.startswith("_"):
                        continue
                    if ext.endswith('py'):
                        ret[name] = os.path.abspath(os.path.join(p, f))
            except:
                self.debug("'%s' is not a valid path" % p)
        return ret

    def _fromNameToFileName(self, lib_name, path=None):
        path = path or self.getMacroPath()[0]
        f_name = lib_name
        if not f_name.endswith('.py'):
            f_name += '.py'

        if os.path.isabs(f_name):
            path, _ = os.path.split(f_name)
            if not path in self.getMacroPath():
                raise Exception("'%s' is not part of the MacroPath" % path)
        else:
            f_name = os.path.join(path, f_name)
        return f_name

    def getOrCreateMacroLib(self, lib_name, macro_name=None):
        """Gets the exiting macro lib or creates a new macro lib file. If
        name is not None, a macro template code for the given macro name is
        appended to the end of the file.

        :param lib_name:
            module name, python file name, or full file name (with path)
        :type lib_name: :obj:`str`
        :param macro_name:
            an optional macro name. If given a macro template code is appended
            to the end of the file (default is None meaning no macro code is
            added)
        :type macro_name: :obj:`str`

        :return:
            a sequence with three items: full_filename, code, line number is 0
            if no macro is created or n representing the first line of code for
            the given macro name.
        :rtype: seq<str, str, int>"""
        # if only given the module name
        try:
            macro_lib = self.getMacroLib(lib_name)
        except UnknownMacroLibrary:
            macro_lib = None

        if macro_name is None:
            line_nb = 0
            if macro_lib is None:
                f_name, code = self.createMacroLib(lib_name), ''
            else:
                f_name = macro_lib.file_path
                f = open(f_name)
                code = f.read()
                f.close()
        else:
            # if given macro name
            if macro_lib is None:
                f_name, code, line_nb = self.createMacro(lib_name, macro_name)
            else:
                macro = macro_lib.get_macro(macro_name)
                if macro is None:
                    f_name, code, line_nb = self.createMacro(
                        lib_name, macro_name)
                else:
                    _, line_nb = macro.code
                    f_name = macro.file_path
                    f = open(f_name)
                    code = f.read()
                    f.close()

        return [f_name, code, line_nb]

    def setMacroLib(self, lib_name, code, auto_reload=True):
        f_name = self._fromNameToFileName(lib_name)
        f = open(f_name, 'w')
        f.write(code)
        f.flush()
        f.close()
        _, name = os.path.split(f_name)
        mod, _ = os.path.splitext(name)
        if auto_reload:
            self.reloadMacroLib(mod)
        return mod

    def createMacroLib(self, lib_name, path=None):
        """Creates a new empty macro library (python module)"""
        f_name = self._fromNameToFileName(lib_name, path)

        if os.path.exists(f_name):
            raise Exception(
                "Unable to create macro lib: '%s' already exists" % f_name)

        f = open(f_name, 'w')
        f.close()
        return f_name

    def createMacro(self, lib_name, macro_name):
        f_name = self._fromNameToFileName(lib_name)

        create = not os.path.exists(f_name)

        template = ''
        if create:
            template += 'from sardana.macroserver.macro import Macro, macro, Type\n\n'
            line_nb = 4
        else:
            template += '\n'
            t = open(f_name, 'rU')
            line_nb = -1
            for line_nb, _ in enumerate(t):
                pass
            line_nb += 3
            t.close()

        f = open(f_name, 'a+')
        f_templ = None
        try:
            dir_name = os.path.realpath(__file__)
            dir_name = os.path.dirname(dir_name)
            template_fname = 'macro_template.txt'
            template_fname = os.path.join(dir_name, template_fname)
            f_templ = open(template_fname, 'r')
            template += f_templ.read()
            f_templ.close()
        except:
            self.debug(
                "Failed to open template macro file. Using simplified template")
            template += MACRO_TEMPLATE
            if f_templ:
                f_templ.close()

        template = template.replace('@macro_name@', macro_name)
        try:
            f.write(template)
            f.flush()
            f.seek(0)
            code = f.read()
        finally:
            f.close()
        return f_name, code, line_nb

    def reloadMacro(self, macro_name, path=None):
        """Reloads the module corresponding to the given macro name

        :raises: MacroServerExceptionList in case the macro is unknown or the
        reload process is not successful

        :param macro_name: macro name
        :param path: a list of absolute path to search for libraries
                     (optional, default=None, means the current MacroPath
                     will be used)"""
        self.reloadMacros([macro_name], path=path)

    def reloadMacros(self, macro_names, path=None):
        """Reloads the modules corresponding to the given macro names

        :raises: MacroServerExceptionList in case the macro(s) are unknown or the
        reload process is not successful

        :param macro_names: a list of macro names
        :param path: a list of absolute path to search for libraries (optional,
                     default=None, means the current MacroPath will be used)"""
        module_names = []
        for macro_name in macro_names:
            module_name = self.getMacro(macro_name).module_name
            module_names.append(module_name)
        self.reloadMacroLibs(module_names, path=path)

    def reloadMacroLibs(self, module_names, path=None):
        """Reloads the given lib(=module) names

        :raises: MacroServerExceptionList in case the reload process is not
        successful for at least one library

        :param module_names: a list of module names
        :param path: a list of absolute path to search for libraries
                     (optional, default=None, means the current MacroPath
                     will be used)"""
        ret = []
        for module_name in module_names:
            m = self.reloadMacroLib(module_name, path=path)
            if m:
                ret.append(m)
        return ret

    def reloadLib(self, module_name, path=None):
        """Reloads the given library(=module) names.

        :raises:
            LibraryError if trying to reload a macro library

        :param module_name: module name
        :param path:
            a list of absolute path to search for libraries [default: None.
            Search in sys.path.]
        :return: the reloaded python module object"""
        if module_name in self._modules:
            raise LibraryError("Cannot use simple " +
                               "reload to reload a Macro Library")

        mod_manager = ModuleManager()
        return mod_manager.reloadModule(module_name, path=None)

    def reloadMacroLib(self, module_name, path=None):
        """Reloads the given library(=module) names.

        :raises:
            LibraryError in case the reload process is not successful

        :param module_name: macro library name (=python module name)
        :param path:
            a list of absolute path to search for libraries [default: None,
            means the current MacroPath will be used]
        :return: the MacroLibrary object for the reloaded macro library"""
        path = path or self.getMacroPath()
        mod_manager = ModuleManager()
        m, exc_info = None, None
        valid, exc_info = mod_manager.isValidModule(module_name, path)
        if not valid:
            params = dict(module=m, name=module_name,
                          macro_server=self.macro_server, exc_info=exc_info)
            return MacroLibrary(**params)

        # if there was previous Macro Library info remove it
        old_macro_lib = self._modules.pop(module_name, None)
        if old_macro_lib is not None:
            for macro in old_macro_lib.get_macros():
                self._macro_dict.pop(macro.name)

        try:
            m = mod_manager.reloadModule(module_name, path)
        except:
            exc_info = sys.exc_info()
        macro_lib = None

        params = dict(module=m, name=module_name,
                      macro_server=self.macro_server, exc_info=exc_info)

        # Dictionary for gathering macros with errors
        macro_errors = {}
        count_correct_macros = 0
        count_incorrect_macros = 0

        if m is None:
            file_name = self._findMacroLibName(module_name)
            if file_name is None:
                if exc_info:
                    msg = format_exception_only_str(*exc_info[:2])
                else:
                    msg = "Error (re)loading macro library '%s'" % module_name
                raise LibraryError(msg, exc_info=exc_info)
            params['file_path'] = file_name
            macro_lib = MacroLibrary(**params)
        else:
            macro_lib = MacroLibrary(**params)
            abs_file = macro_lib.file_path
            _is_macro = functools.partial(is_macro, abs_file=abs_file,
                                          logger=self)
            for _, macro in inspect.getmembers(m, _is_macro):
                try:
                    isoverwritten = False
                    macro_name = macro.__name__
                    if macro_name in self._overwritten_macros:
                        isoverwritten = True
                    elif (macro_name in list(self._macro_dict.keys())
                            and self._macro_dict[macro_name].lib != macro_lib):
                        isoverwritten = True
                        msg = ('Macro "{0}" defined in "{1}" macro library'
                               + ' has been overwritten by "{2}" macro library'
                               )
                        old_lib_name = self._macro_dict[macro_name].lib.name
                        self.debug(msg.format(macro_name, old_lib_name,
                                              macro_lib.name))
                        self._overwritten_macros.append(macro_name)

                    self.addMacro(macro_lib, macro, isoverwritten)
                    count_correct_macros += 1
                except Exception as e:
                    count_incorrect_macros += 1
                    self.error("Error adding macro %s", macro.__name__)
                    self.debug("Details:", exc_info=1)
                    macro_errors[macro.__name__] = str(e)
        try:
            if macro_lib.has_macros():
                self._modules[module_name] = macro_lib
            return macro_lib
        finally:
            if macro_errors:
                msg = ""
                for key, value in macro_errors.items():
                    msg_part = ("\n" + "Error adding macro(s): " + key + "\n"
                                + "It presents an error: \n" + str(value))
                    msg += str(msg_part) + "\n"
                correct_macros = ("%d macro(s) correctly loaded" %
                                  count_correct_macros)
                incorrect_macros = ("%d macro(s) could not be loaded" %
                                    count_incorrect_macros)

                msg = (msg + "\n" + "Summary:" + "\n" + correct_macros
                       + "\n" + incorrect_macros + "\n")

                if count_correct_macros == 0:
                    msg += "\nUse addmaclib to reload the corrected macro(s)\n"
                if count_correct_macros != 0:
                    msg += "\nUse relmaclib to reload the corrected macro(s)\n"
                raise Exception(msg)

    def addMacro(self, macro_lib, macro, isoverwritten=False):
        add = self.addMacroFunction
        if inspect.isclass(macro):
            add = self.addMacroClass
        return add(macro_lib, macro, isoverwritten)

    def addMacroClass(self, macro_lib, klass, isoverwritten=False):
        macro_name = klass.__name__
        action = (macro_lib.has_macro(macro_name) and "Updating") or "Adding"
        self.debug("%s macro class %s" % (action, macro_name))

        params = dict(macro_server=self.macro_server, lib=macro_lib,
                      klass=klass, isoverwritten=isoverwritten)
        macro_class = MacroClass(**params)
        macro_lib.add_macro_class(macro_class)
        self._macro_dict[macro_name] = macro_class

    def addMacroFunction(self, macro_lib, func, isoverwritten=False):
        macro_name = func.__name__
        action = (macro_lib.has_macro(macro_name) and "Updating") or "Adding"
        self.debug("%s macro function %s" % (action, macro_name))

        params = dict(macro_server=self.macro_server, lib=macro_lib,
                      function=func, isoverwritten=isoverwritten)
        macro_function = MacroFunction(**params)
        macro_lib.add_macro_function(macro_function)
        self._macro_dict[macro_name] = macro_function

    def getMacroLibNames(self):
        return sorted(self._modules.keys())

    def getMacroLibs(self, filter=None):
        if filter is None:
            return self._modules
        expr = re.compile(filter, re.IGNORECASE)
        ret = {}
        for name, macro_lib in self._modules.items():
            if expr.match(name) is None:
                continue
            ret[name] = macro_lib
        return ret

    def getMacros(self, filter=None):
        """Returns a :obj:`dict` containing information about macros.

        :param filter:
            a regular expression for macro names [default: None, meaning all
            macros]
        :type filter: :obj:`str`
        :return: a :obj:`dict` containing information about macros
        :rtype:
            :obj:`dict`\<:obj:`str`\, :class:`~sardana.macroserver.msmetamacro.MacroCode`\>"""
        if filter is None:
            return self._macro_dict
        expr = re.compile(filter, re.IGNORECASE)

        ret = {}
        for name, macro in self._macro_dict.items():
            if expr.match(name) is None:
                continue
            ret[name] = macro
        return ret

    def getMacroClasses(self, filter=None):
        """Returns a :obj:`dict` containing information about macro classes.

        :param filter:
            a regular expression for macro names [default: None, meaning all
            macros]
        :type filter: :obj:`str`
        :return: a :obj:`dict` containing information about macro classes
        :rtype:
            :obj:`dict`\<:obj:`str`\, :class:`~sardana.macroserver.msmetamacro.MacroClass`\>"""
        macros = self.getMacros(filter=filter)
        macro_classes = {}
        for name, macro in list(macros.items()):
            if macro.get_type() == ElementType.MacroClass:
                macro_classes[name] = macro
        return macro_classes

    def getMacroFunctions(self, filter=None):
        """Returns a :obj:`dict` containing information about macro functions.

        :param filter:
            a regular expression for macro names [default: None, meaning all
            macros]
        :type filter: :obj:`str`
        :return: a :obj:`dict` containing information about macro functions
        :rtype:
            :obj:`dict`\<:obj:`str`\, :class:`~sardana.macroserver.msmetamacro.MacroFunction`\>"""
        macros = self.getMacros(filter=filter)
        macro_classes = {}
        for name, macro in list(macros.items()):
            if macro.get_type() == ElementType.MacroFunction:
                macro_classes[name] = macro
        return macro_classes

    def getMacroNames(self):
        return sorted(self._macro_dict.keys())

    def getMacro(self, macro_name):
        ret = self._macro_dict.get(macro_name)
        if ret is None:
            raise UnknownMacro("Unknown macro %s" % macro_name)
        return ret

    def getMacroClass(self, macro_name):
        return self.getMacro(macro_name)

    def getMacroFunction(self, macro_name):
        return self.getMacro(macro_name)

    def removeMacro(self, macro_name):
        self._macro_dict.pop(macro_name)

    def getMacroLib(self, name):
        if os.path.isabs(name):
            abs_file_name = name
            for lib in list(self._modules.values()):
                if lib.file_path == abs_file_name:
                    return lib
        elif name.count(os.path.extsep):
            file_name = name
            for lib in list(self._modules.values()):
                if lib.file_name == file_name:
                    return lib
        module_name = name
        ret = self._modules.get(module_name)
        if ret is None:
            raise UnknownMacroLibrary("Unknown macro library %s" % module_name)
        return ret

    def getMacroCode(self, macro_name):
        return self.getMacro(macro_name).code_object

    def getMacroClassCode(self, macro_name):
        return self.getMacroClass(macro_name).klass

    def getMacroFunctionCode(self, macro_name):
        return self.getMacroFunction(macro_name).function

    def getMacroInfo(self, macro_names, format='json'):
        if isinstance(macro_names, str):
            macro_names = [macro_names]
        ret = []
        json_codec = CodecFactory().getCodec(format)
        for macro_name in macro_names:
            macro_meta = self.getMacro(macro_name)
            ret.append(json_codec.encode(('', macro_meta.serialize()))[1])
        return ret

    def _createMacroNode(self, macro_name, macro_params_raw):
        macro = self.getMacro(macro_name)
        params_def = macro.get_parameter()
        macro_params_str = " ".join(macro_params_raw)
        param_parser = ParamParser(params_def)
        # parse string with macro params to the correct list representation
        macro_params = param_parser.parse(macro_params_str)
        return createMacroNode(macro_name, params_def, macro_params)

    def decodeMacroParameters(self, door, raw_params):
        """Decode macro parameters

        :param door: (sardana.macroserver.msdoor.MSDoor) door object
        :param raw_params: (lxml.etree._Element or list) xml element
            representing macro with subelements representing parameters or
            list with macro name followed by parameter values
        """
        if isinstance(raw_params, etree._Element):
            macro_name = raw_params.get("name")
        elif isinstance(raw_params, list):
            # leave only macro parameters in the list
            macro_name = raw_params.pop(0)
        else:
            raise Exception("Wrong format of raw_params object")
        macro_meta = self.getMacro(macro_name)
        params_def = macro_meta.get_parameter()
        type_manager = door.type_manager
        try:
            out_par_list = ParamDecoder(type_manager, params_def, raw_params)
        except WrongParam as out_e:
            # only if raw params are passed as a list e.g. using macro API
            # execMacro("mv", mot01, 0.0) and parameters definition allows to
            # decode it from a flat list we give it a try
            if (is_flat_list(raw_params) and
                    FlatParamDecoder.isPossible(params_def)):
                self.debug("Trying flat parameter decoder due to: %s" % out_e)
                try:
                    out_par_list = FlatParamDecoder(type_manager, params_def,
                                                    raw_params)
                except WrongParam as in_e:
                    msg = ("Either of: %s or %s made it impossible to decode"
                           " parameters" % (out_e, in_e))
                    raise WrongParam(msg)
            else:
                raise out_e
        return macro_meta, raw_params, out_par_list

    def strMacroParamValues(self, par_list):
        """strMacroParamValues(list<string> par_list) -> list<string>

           Creates a short string representantion of the parameter values list.
           Params:
               - par_list: list of strings representing the parameter values.
           Return:
               a list containning an abreviated version of the par_list argument.
        """
        ret = []
        for p in par_list:
            param_str = str(p)
            if len(param_str) > 9:
                param_str = param_str[:9] + "..."
            ret.append(param_str)
        return ret

    def prepareMacro(self, macro_class, par_list,
                     init_opts={}, prepare_opts={}):
        """Creates the macro object and calls its prepare method.
           The return value is a tuple (MacroObject, return value of prepare)
        """
        macro = self.createMacroObj(macro_class, par_list, init_opts=init_opts)
        prepare_result = self.prepareMacroObj(
            macro, par_list, prepare_opts=prepare_opts)
        return macro, prepare_result

    def createMacroObj(self, macro_class, par_list, init_opts={}):
        macro_env = macro_class.env
        macro_name = macro_class.__name__

        environment = init_opts.get('environment')
        executor = init_opts.get('executor')
        door_name = executor.door.name

        r = []
        for env in macro_env:
            if not environment.has_env(env,
                                       macro_name=macro_name,
                                       door_name=door_name):
                r.append(env)
        if r:
            raise MissingEnv("The macro %s requires the following missing "
                             "environment to be defined: %s"
                             % (macro_name, str(r)))

        macro_opts = {
            'no_exec': True,
            'create_thread': True,
            'external_prepare': True
        }

        macro_opts.update(init_opts)
        macroObj = macro_class(*par_list, **macro_opts)
        return macroObj

    def prepareMacroObj(self, macro, par_list, prepare_opts={}):
        return macro.prepare(*par_list, **prepare_opts)

    def createMacroObjFromMeta(self, meta, par_list, init_opts={}):
        code = meta.code_object
        macro_env = code.env or ()

        environment = init_opts.get('environment')
        executor = init_opts.get('executor')
        door_name = executor.door.name
        macro_name = meta.name
        r = []
        for env in macro_env:
            if not environment.has_env(env,
                                       macro_name=macro_name,
                                       door_name=door_name):
                r.append(env)
        if r:
            raise MissingEnv("The macro %s requires the following missing "
                             "environment to be defined: %s"
                             % (macro_name, str(r)))

        macro_opts = dict(no_exec=True, create_thread=True,
                          external_prepare=True)

        macro_opts.update(init_opts)
        if meta.get_type() == ElementType.MacroClass:
            macroObj = meta.macro_class(*par_list, **macro_opts)
        else:
            macro_opts['function'] = code
            macroObj = MacroFunc(*par_list, **macro_opts)
        return macroObj

    def getMacroExecutor(self, door):
        me = self._macro_executors.get(door)
        if me is None:
            self._macro_executors[door] = me = MacroExecutor(door)
        return me


class LogMacroFilter(logging.Filter):

    def __init__(self, param=None):
        self.param = param

    def filter(self, record):
        allow = True
        if record.levelname == "DEBUG":
            if not isinstance(record.msg, str):
                allow = False
                return allow
            if record.msg.find("[START]") != -1:
                msg = record.msg
                start = msg.index("'") + 1
                end = msg.index("->", start)
                msg = msg[start:end]
                msg = msg.replace("(", " ").replace(")", "").replace(
                    "[", "").replace("]", "")
                msg = msg.replace(", ", " ")
                msg = msg.replace(",", " ")
                msg = msg.replace(".*", "")
                while msg.find("  ") != -1:
                    msg = msg.replace("  ", " ")
                if msg[0] == "_":
                    allow = False
                else:
                    msg_split = msg.split(" ")
                    msg = ""
                    for i in range(0, len(msg_split)):
                        if msg_split[i].find(" ") == -1:
                            msg_split[i] = msg_split[i].replace("'", " ")
                        msg = msg + " " + str(msg_split[i])
                    while msg.find("  ") != -1:
                        msg = msg.replace("  ", " ")
                    record.msg = "\n-- " + time.ctime() + "\n" + msg
                    allow = True
            else:
                allow = False
        return allow


class LogMacroManager(Logger):

    """Manage user-oriented macro logging to a file. It is configurable with
    LogMacro, LogMacroMode, LogMacroFormat and LogMacroDir environment
    variables.

    .. note::
        The LogMacroManager class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including its removal) may occur if
        deemed necessary by the core developers.
    """

    DEFAULT_DIR = os.path.join(os.sep, "tmp")
    DEFAULT_FMT = "%(message)s"
    DEFAULT_MODE = 0

    def __init__(self, macro_obj):
        name = macro_obj.getName() + ".LogMacroManager"
        Logger.__init__(self, name)
        self._macro_obj = macro_obj
        self._file_handler = None
        self._enabled = False

    def getFilterClass(self):
        """Get filter class.

        First look if a custom filter class was set using
        sardanacustomsettings.LOG_MACRO_FILTER variable, if not, silently
        return None (no filter will be applied).

        If the custom filter class was incorrectly set or class is not
        importable warn the user and return None (no filter will be applied).
        """
        filter_class = None
        from sardana import sardanacustomsettings
        try:
            log_macro_filter = getattr(sardanacustomsettings,
                                       "LOG_MACRO_FILTER")
        except AttributeError:
            pass
        else:
            if isinstance(log_macro_filter, str):
                try:
                    module_name, filter_name = log_macro_filter.rsplit('.', 1)
                    __import__(module_name)
                    module = sys.modules[module_name]
                    filter_class = getattr(module, filter_name)
                except Exception:
                    msg = "sardanacustomsettings.LOG_MACRO_FILTER has wrong" \
                          " format or class is not importable." \
                          " No filter will be used."
                    self.warning(msg)
                    self.debug(exc_info=True)
        return filter_class

    def enable(self):
        """Enable macro logging only if the following requirements are
        fulfilled:
            * this is the top-most macro
            * macro logging is enabled by user

        :return: True or False, depending if logging was enabled or not
        :rtype: boolean
        """
        macro_obj = self._macro_obj
        executor = macro_obj.executor
        door = macro_obj.door

        # enable logging only for the top-most macros
        if macro_obj.getParentMacro() is not None:
            return False
        # enable logging only if configured by user
        try:
            enabled = macro_obj.getEnv("LogMacro")
        except UnknownEnv:
            return False
        if not enabled:
            return False

        try:
            logging_mode = macro_obj.getEnv("LogMacroMode")
        except UnknownEnv:
            logging_mode = self.DEFAULT_MODE
        try:
            logging_path = macro_obj.getEnv("LogMacroDir")
        except UnknownEnv:
            logging_path = self.DEFAULT_DIR
            macro_obj.setEnv("LogMacroDir", logging_path)

        door_name = door.name
        # Cleaning name in case alias does not exist
        door_name = door_name.replace(":", "_").replace("/", "_")
        file_name = "session_" + door_name + ".log"
        log_file = os.path.join(logging_path, file_name)

        if logging_mode:
            bck_counts = 100
        else:
            bck_counts = 0

        self._file_handler = file_handler = \
            logging.handlers.RotatingFileHandler(log_file,
                                                 backupCount=bck_counts)
        file_handler.doRollover()

        filter_class = self.getFilterClass()
        if filter_class is not None:
            try:
                filter_ = filter_class()
            except Exception:
                msg = "Not possible to instantiate %s class. No filter will" \
                      " be used." % filter_class
                self.warning(msg)
                self.debug(exc_info=True)
            else:
                file_handler.addFilter(filter_)

        try:
            format_to_set = macro_obj.getEnv("LogMacroFormat")
        except UnknownEnv:
            format_to_set = self.DEFAULT_FMT
        log_format = logging.Formatter(format_to_set)
        file_handler.setFormatter(log_format)
        # attach the same handler to two different loggers due to
        # lack of hierarchy between them (see: sardana-org/sardana#703)
        macro_obj.addLogHandler(file_handler)
        executor.addLogHandler(file_handler)
        self._enabled = True
        return True

    def disable(self):
        """Disable macro logging only if it was enabled before.

        :return: True or False, depending if logging was disabled or not
        :rtype: boolean
        """

        if not self._enabled:
            return False
        macro_obj = self._macro_obj
        executor = macro_obj.executor
        file_handler = self._file_handler
        macro_obj.removeLogHandler(file_handler)
        executor.removeLogHandler(file_handler)

        return True


class MacroExecutor(Logger):

    """ """

    class RunSubXMLHook:

        def __init__(self, me, xml):
            self._me = me
            self._xml = xml

        def __call__(self):
            self._me._runXMLMacro(xml=self._xml)

    def __init__(self, door):
        self._door = door
        self._macro_counter = 0

        # dict<PoolElement, set<Macro>>
        # key PoolElement - reserved object
        # value set<Macro> macros that reserved the object
        self._reserved_objs = {}

        # dict<Macro, seq<PoolElement>>
        # key Macro - macro object
        # value - sequence of reserverd objects by the macro
        self._reserved_macro_objs = {}
        # dict<Macro, seq<PoolElement>>
        # key Macro - macro object
        # value - sequence of reserverd objects by the macro
        #   which were already successfully stopped
        self._stopped_macro_objs = {}

        # reset the stacks
#        self._macro_stack = None
#        self._xml_stack = None
        self._macro_stack = []
        self._xml_stack = []
        self._macro_pointer = None
        self._abort_thread = None
        self._aborted = False
        self._stop_thread = None
        self._stopped = False
        self._paused = False
        self._released = False
        self._last_macro_status = None
        # threading events for synchronization of stopping/abortting of
        # reserved objects
        self._stop_done = None
        self._abort_done = None

        name = "%s.%s" % (str(door), self.__class__.__name__)
        self._macro_status_codec = CodecFactory().getCodec('json')
        self.call__init__(Logger, name)

    def getDoor(self):
        return self._door

    door = property(getDoor)

    def getMacroServer(self):
        return self.door.macro_server

    macro_server = property(getMacroServer)

    @property
    def macro_manager(self):
        return self.macro_server.macro_manager

    def getGeneralHooks(self):
        """Get data structure containing definition of the general hooks.

        .. note::
        The `general_hooks` has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including its removal) may occur if
        deemed necessary by the core developers.

        .. todo::
        General hooks should reflect the state of configuration at the
        macro/sequence start.
        """

        try:
            return self.door.get_env("_GeneralHooks")["_GeneralHooks"]
        except KeyError:
            return []

    general_hooks = property(getGeneralHooks)

    def getNewMacroID(self):
        self._macro_counter -= 1
        return self._macro_counter

    def _createMacroNode(self, macro_name, macro_params):
        return self.macro_manager._createMacroNode(macro_name, macro_params)

    def _preprocessParameters(self, par_str_list):
        if not par_str_list[0].lstrip().startswith('<'):
            xml_root = xml_seq = etree.Element('sequence')
            macro_name = par_str_list[0]
            macro_params = par_str_list[1:]

            def quote_string(string):
                # if string contains double quotes, use single quotes,
                # otherwise use double quotes
                if re.search('"', string):
                    return "'{}'".format(string)
                else:
                    return '"{}"'.format(string)

            # param parser relies on whitespace separation of parameter values
            # quote string parameter values containing whitespaces
            macro_params_quoted = []
            for param in macro_params:
                if (not re.match(r".*\s+.*", param)  # no white spaces
                        or re.match(r"^'.*\s+.*'$", param)  # already quoted
                        or re.match(r'^".*\s+.*"$', param)):  # already quoted
                    macro_params_quoted.append(param)
                else:
                    macro_params_quoted.append(quote_string(param))
            macro_node = self._createMacroNode(macro_name, macro_params_quoted)
            xml_macro = macro_node.toXml()
            xml_seq.append(xml_macro)
        else:
            xml_root = etree.fromstring(par_str_list[0])

        macro_nodes = xml_root.findall('.//macro')

        # make sure macros exist
        self.__checkXMLSequence(macro_nodes)

        # fill the xml with macro id macro_line
        self.__fillXMLSequence(macro_nodes)

        return xml_root

    def __checkXMLSequence(self, macros):
        for macro in macros:
            name = macro.get('name')
            self.macro_manager.getMacro(name)

    def __fillXMLSequence(self, macros):
        for macro in macros:
            eid = macro.get('id')
            if eid is None:
                eid = str(self.getNewMacroID())
                macro.set('id', eid)

    def __preprocessResult(self, result):
        """decodes the given output from a macro in order to be able to send to
        the result channel as a sequence<str>

        :param out: output value

        :return: the output as a sequence of strings
        """
        if result is None:
            return ()
        if is_non_str_seq(result):
            result = list(map(str, result))
        else:
            result = (str(result),)
        return result

    def _decodeMacroParameters(self, params):
        return self.macro_manager.decodeMacroParameters(self.door, params)

    def _composeMacroLine(self, macro_name, macro_params, macro_id):
        # recursive map to maintain the list objects structure
        params_str_list = recur_map(str, macro_params)
        # plain map to be able to perform join (only strings may be joined)
        params_str_list = list(map(str, params_str_list))
        params_str = ', '.join(params_str_list)
        macro_id = macro_id
        # create macro_line - string representation of macro, its parameters
        # and id
        macro_line = "%s(%s) -> %s" % (macro_name, params_str, macro_id)
        return macro_line

    def _prepareGeneralHooks(self, macro_obj):
        if not isinstance(macro_obj, Hookable):
            return
        general_hooks = self.general_hooks
        if len(general_hooks) == 0:
            return
        for hook_info_raw, hook_places in general_hooks:
            hook_info_tokens = hook_info_raw.split(" ", 1)
            hook_name = hook_info_tokens[0]
            hook_info = [hook_name]
            if len(hook_info_tokens) == 2:
                hook_params_raw = hook_info_tokens[1]
                hook_param_def = self.macro_manager.getMacro(
                    hook_name).get_parameter()
                param_parser = ParamParser(hook_param_def)
                hook_params = param_parser.parse(hook_params_raw)
                hook_info += hook_params
            hook = ExecMacroHook(macro_obj, hook_info)
            macro_obj.appendHook((hook, hook_places))

    def _prepareXMLMacro(self, xml_macro, parent_macro=None):
        macro_meta, _, macro_params = self._decodeMacroParameters(xml_macro)
        macro_name = macro_meta.name
        macro_id = xml_macro.get("id")
        macro_line = self._composeMacroLine(macro_name, macro_params, macro_id)
        init_opts = {
            'id': macro_id,
            'macro_line': macro_line,
            'parent_macro': parent_macro,
        }

        macro_obj = self._createMacroObj(macro_meta, macro_params, init_opts)

        self._prepareGeneralHooks(macro_obj)

        for macro in xml_macro.findall('macro'):
            hook = MacroExecutor.RunSubXMLHook(self, macro)
            hook_hints = macro.findall('hookPlace')
            if hook_hints is None:
                macro_obj.appendHook((hook, []))
            else:
                hook_places = [h.text for h in hook_hints]
                macro_obj.appendHook((hook, hook_places))

        prepare_result = self._prepareMacroObj(macro_obj, macro_params)
        return macro_obj, prepare_result

    def _createMacroObj(self, macro_name_or_meta, pars, init_opts={}):
        macro_meta = macro_name_or_meta
        if isinstance(macro_meta, str):
            macro_meta = self.macro_manager.getMacro(macro_meta)

        macro_opts = {
            'executor': self,
            'environment': self.macro_server
        }
        macro_opts.update(init_opts)
        if 'id' not in macro_opts:
            macro_opts['id'] = str(self.getNewMacroID())

        macroObj = self.macro_manager.createMacroObjFromMeta(macro_meta, pars,
                                                             init_opts=macro_opts)

        return macroObj

    def _prepareMacroObj(self, macro_obj, pars, prepare_opts={}):
        return self.macro_manager.prepareMacroObj(macro_obj, pars,
                                                  prepare_opts=prepare_opts)

    def prepareMacroObj(self, macro_name_or_meta, pars, init_opts={},
                        prepare_opts={}):
        """Prepare a new macro for execution

        :param macro_name_or_meta name: name of the macro to be prepared or
                                        the macro meta itself
        :param pars: list of parameter objects
        :param init_opts: keyword parameters for the macro constructor
        :param prepare_opts: keyword parameters for the macro prepare

        :return: a tuple of two elements: macro object, the result of preparing the macro"""
        macroObj = self._createMacroObj(
            macro_name_or_meta, pars, init_opts=init_opts)
        prepare_result = self._prepareMacroObj(
            macroObj, pars, prepare_opts=prepare_opts)
        return macroObj, prepare_result

    def prepareMacro(self, pars, init_opts={}, prepare_opts={}):
        """Prepare a new macro for execution
           Several different parameter formats are supported:
           1. several parameters:
             1.1 executor.prepareMacro('ascan', 'th', '0', '100', '10', '1.0')
                 executor.prepareMacro('mv', [['th', '0']])
                 executor.prepareMacro('mv', 'th', '0') # backwards compatibility - see note
             1.2 executor.prepareMacro('ascan', 'th', 0, 100, 10, 1.0)
                 executor.prepareMacro('mv', [['th', 0]])
                 executor.prepareMacro('mv', 'th', 0) # backwards compatibility - see note
             1.3 th = self.getObj('th');
                 executor.prepareMacro('ascan', th, 0, 100, 10, 1.0)
                 executor.prepareMacro('mv', [[th, 0]])
                 executor.prepareMacro('mv', th, 0) # backwards compatibility - see note
           2. a sequence of parameters:
              2.1 executor.prepareMacro(['ascan', 'th', '0', '100', '10', '1.0')
                  executor.prepareMacro(['mv', [['th', '0']]])
                  executor.prepareMacro(['mv', 'th', '0']) # backwards compatibility - see note
              2.2 executor.prepareMacro(('ascan', 'th', 0, 100, 10, 1.0))
                  executor.prepareMacro(['mv', [['th', 0]]])
                  executor.prepareMacro(['mv', 'th', 0]) # backwards compatibility - see note
              2.3 th = self.getObj('th');
                  executor.prepareMacro(['ascan', th, 0, 100, 10, 1.0])
                  executor.prepareMacro(['mv', [[th, 0]]])
                  executor.prepareMacro(['mv', th, 0]) # backwards compatibility - see note
           3. a space separated string of parameters (this is not compatible
              with multiple or nested repeat parameters, furthermore the repeat
              parameter must be the last one):
              executor.prepareMacro('ascan th 0 100 10 1.0')
              executor.prepareMacro('mv %s 0' % motor.getName())

        .. note:: From Sardana 2.0 the repeat parameter values must be passed
            as lists of items. An item of a repeat parameter containing more
            than one member is a list. In case when a macro defines only one
            repeat parameter and it is the last parameter, for the backwards
            compatibility reasons, the plain list of items' members is allowed.

        :param pars: the command parameters as explained above
        :param opts: keyword optional parameters for prepare
        :return: a tuple of two elements: macro object, the result of preparing the macro
        """
        par0 = pars[0]
        if len(pars) == 1:
            if is_pure_str(par0):
                # dealing with sth like args = ('ascan th 0 100 10 1.0',)
                pars = par0.split()
                macro_name, macro_params = pars[0], pars[1:]
                macro_node = self._createMacroNode(macro_name, macro_params)
                pars = macro_node.toList()
            elif is_non_str_seq(par0):
                # dealing with sth like args = (['ascan', 'th', '0', '100', '10', '1.0'],)
                # or args = (['mv', [[mot01, 0], [mot02, 0]]])
                pars = par0
        # dealing with sth like args = ('ascan', 'th', '0', '100', '10', '1.0')
        # or args = ('mv', [[mot01, 0], [mot02, 0]])

        # in case parameters were passed as objects cast them to strings
        # but maintain None's to be able to discover missing params
        pars = recur_map(str, pars, keep_none=True)

        meta_macro, _, macro_params = self._decodeMacroParameters(pars)
        macro_name = meta_macro.name
        macro_id = init_opts.get("id")
        if macro_id is None:
            init_opts["id"] = macro_id
        macro_line = self._composeMacroLine(macro_name,
                                            macro_params,
                                            macro_id)

        init_opts['macro_line'] = macro_line

        macro_obj, prepare_result = self.prepareMacroObj(meta_macro,
                                                         macro_params,
                                                         init_opts,
                                                         prepare_opts)

        self._prepareGeneralHooks(macro_obj)

        return macro_obj, prepare_result

    def getRunningMacro(self):
        return self._macro_pointer

    def clearRunningMacro(self):
        """Clear pointer to the running macro.

        ..warning:: Do not call it while the macro is running.
        """
        self._macro_pointer = None

    def __stopObjects(self):
        """Stops all the reserved objects in the executor"""
        for macro, objs in list(self._reserved_macro_objs.items()):
            if self._aborted:
                break  # someone aborted, no sense to stop anymore
            self._stopped_macro_objs[macro] = stopped_macro_objs = []
            for obj in objs:
                if self._aborted:
                    break  # someone aborted, no sense to stop anymore
                self.output(
                    "Stopping {} reserved by {}".format(obj, macro._name))
                try:
                    obj.stop()
                except AttributeError:
                    pass
                except:
                    self.warning("Unable to stop %s" % obj)
                    self.debug("Details:", exc_info=1)
                else:
                    self.output("{} stopped".format(obj))
                    stopped_macro_objs.append(obj)

    def __abortObjects(self):
        """Aborts all the reserved objects in the executor"""
        for macro, objs in list(self._reserved_macro_objs.items()):
            stopped_macro_objs = self._stopped_macro_objs[macro]
            for obj in objs:
                if obj in stopped_macro_objs:
                    continue
                self.output(
                    "Aborting {} reserved by {}".format(obj, macro._name))
                try:
                    obj.abort()
                except AttributeError:
                    pass
                except:
                    self.warning("Unable to abort %s" % obj)
                    self.debug("Details:", exc_info=1)
                else:
                    self.output("{} aborted".format(obj))

    def _setStopDone(self, _):
        self._stop_done.set()

    def _waitStopDone(self, timeout=None):
        self._stop_done.wait(timeout)

    def _isStopDone(self):
        return self._stop_done.is_set()

    def _setAbortDone(self, _):
        self._abort_done.set()

    def _waitAbortDone(self, timeout=None):
        self._abort_done.wait(timeout)

    def _isAbortDone(self):
        return self._abort_done.is_set()

    def abort(self):
        """**Internal method**. Aborts the macro abruptly."""
        # carefull: Inside this method never call a method that has the
        # mAPI decorator
        self._aborted = True
        if not self._isStopDone():
            Logger.debug(self, "Break stopping...")
            raise_in_thread(ReleaseException, self._stop_thread)
        self.macro_server.add_job(self._abort, self._setAbortDone)

    def release(self):
        """**Internal method**. Release the macro from hang situations

        Hanged situations:
        * hanged process of aborting reserved objects
        * hanged macro on_abort method.
        """
        # carefull: Inside this method never call a method that has the
        # mAPI decorator
        self._released = True
        if self._isAbortDone():
            m = self.getRunningMacro()
            Logger.debug(self, "Break {}.on_abort...".format(m._name))
            raise_in_thread(ReleaseException, m._macro_thread)
        else:
            Logger.debug(self, "Break aborting...")
            raise_in_thread(ReleaseException, self._abort_thread)


    def stop(self):
        self._stopped = True
        self.macro_server.add_job(self._stop, self._setStopDone)

    def _abort(self):
        self._abort_thread = threading.current_thread()
        if self._stopped:
            # stopping did not finish on its own - we are aborting it
            # but need to wait anyway so its thread finishes
            self._waitStopDone()
        m = self.getRunningMacro()
        if m is not None:
            m.abort()
        self.__abortObjects()

    def _stop(self):
        self._stop_thread = threading.current_thread()
        m = self.getRunningMacro()
        if m is not None:
            m.stop()
            if m.isPaused():
                m.resume(cb=self._macroResumed)
        self.__stopObjects()

    def pause(self):
        self._paused = True
        m = self.getRunningMacro()
        if m is not None:
            m.pause(cb=self._macroPaused)

    def _macroPaused(self, m):
        """Calback that is executed when the macro has efectively paused"""
        self.sendMacroStatusPause()
        self.sendState(Macro.Pause)

    def resume(self):
        if not self._paused:
            return
        self._paused = False
        m = self.getRunningMacro()
        if m is not None:
            m.resume(cb=self._macroResumed)

    def _macroResumed(self, m):
        """Callback that is executed when the macro has effectively resumed
        execution after being paused"""
        self.sendMacroStatusResume()
        self.sendState(Macro.Running)

    def run(self, params, asynch=True):
        """Runs the given macro(s)

        :param params: (sequence<str>) can be either a sequence of <macro name> [, <macro_parameter> ]
                       or a sequence with a single element which represents the xml string for a
                       macro script
        :return: (lxml.etree.Element) the xml representation of the running macro
        """
        # dict<PoolElement, set<Macro>>
        # key PoolElement - reserved object
        # value set<Macro> macros that reserved the object
        self._reserved_objs = {}

        # dict<Macro, seq<PoolElement>>
        # key Macro - macro object
        # value - sequence of reserved objects by the macro
        self._reserved_macro_objs = {}

        # reset the stacks
        self._macro_stack = []
        self._xml_stack = []
        self._macro_pointer = None
        self._stop_done = threading.Event()
        self._abort_done = threading.Event()
        self._aborted = False
        self._stopped = False
        self._paused = False
        self._last_macro_status = None

        # convert given parameters into an xml
        self._xml = self._preprocessParameters(params)

        if asynch:
            # start the job of actually running the macro
            self.macro_server.add_job(self.__runXML, self._jobEnded)
            # return the proper xml
            return self._xml
        else:
            self.__runXML()
            # return self._macro_pointer.getResult()

    def _jobEnded(self, *args, **kw):
        self.debug("Job ended (stopped=%s, aborted=%s)",
                   self._stopped, self._aborted)

    def __runXML(self):
        self.sendState(Macro.Running)
        try:
            self.__runStatelessXML()
            self.sendState(Macro.Finished)
        except (StopException, AbortException):
            self.sendState(Macro.Abort)
        except Exception:
            self.sendState(Macro.Exception)
        finally:
            self._macro_stack = None
            self._xml_stack = None

    def __runStatelessXML(self, xml=None):
        if xml is None:
            xml = self._xml
        node = xml.tag
        if node == 'sequence':
            for xml_macro in xml.findall('macro'):
                self.__runXMLMacro(xml_macro)
        elif node == 'macro':
            self.__runXMLMacro(xml)

    def __runXMLMacro(self, xml):
        assert xml.tag == 'macro'
        try:
            macro_obj, _ = self._prepareXMLMacro(xml)
        except AbortException as ae:
            raise ae
        except Exception as e:
            door = self.door
            door.error("Error: %s", str(e))
            door.debug("Error details:", exc_info=1)
            raise e

        self._xml_stack.append(xml)
        try:
            self.runMacro(macro_obj)
        finally:
            self._xml_stack.pop()

    _runXMLMacro = __runXMLMacro

    def runMacro(self, macro_obj):

        name = macro_obj._getName()
        desc = macro_obj._getDescription()
        door = self.door

        log_macro_manager = LogMacroManager(macro_obj)
        log_macro_manager.enable()

        if self._aborted:
            self.sendMacroStatusAbort()
            raise AbortException("aborted between macros (before %s)" % name)
        elif self._stopped:
            self.sendMacroStatusStop()
            raise StopException("stopped between macros (before %s)" % name)
        macro_exp, tb, result = None, None, None
        try:
            self.debug("[START] runMacro %s" % desc)
            self._macro_pointer = macro_obj
            self._macro_stack.append(macro_obj)
            for step in macro_obj.exec_():
                self.sendMacroStatus((step,))
            result = macro_obj.getResult()
            # sending result only if we are the top most macro
            if macro_obj.hasResult() and macro_obj.getParentMacro() is None:
                result_repr = self.__preprocessResult(result)
                door.debug("sending result %s", result_repr)
                self.sendResult(result_repr)
        except AbortException as ae:
            macro_exp = ae
        except StopException as se:
            macro_exp = se
        except MacroServerException as mse:
            exc_info = sys.exc_info()
            macro_exp = mse
            if not mse.traceback:
                mse.traceback = traceback.format_exc()
        except DevFailed as df:
            exc_info = sys.exc_info()
            exp_pars = {'type': df.args[0].reason,
                        'msg': df.args[0].desc,
                        'args': df.args,
                        'traceback': traceback.format_exc()}
            macro_exp = MacroServerException(exp_pars)
        except Exception as err:
            exc_info = sys.exc_info()
            exp_pars = {'type': err.__class__.__name__,
                        'msg': str(err),
                        'args': err.args,
                        'traceback': traceback.format_exc()}
            macro_exp = MacroServerException(exp_pars)

        # make sure the macro's on_abort is called and that a proper macro
        # status is sent
        if self._aborted:
            self._waitAbortDone()
            self.output("Executing {}.on_abort method...".format(name))
            macro_obj._abortOnError()
            self.sendMacroStatusAbort()
        elif self._stopped:
            self._waitStopDone()
            self.output("Executing {}.on_stop method...".format(name))
            macro_obj._stopOnError()
            self.sendMacroStatusStop()

        self.returnObjs(self._macro_pointer)

        # From this point on don't call any method of macro_obj which is part
        # of the mAPI (methods decorated with @mAPI) to avoid throwing an
        # AbortException if an Abort has been performed.
        if macro_exp is not None:
            if not self._stopped and not self._aborted:
                self.sendMacroStatusException(exc_info)
            self.debug("[ENDEX] (%s) runMacro %s" %
                       (macro_exp.__class__.__name__, name))
            if isinstance(macro_exp, MacroServerException):
                if macro_obj.parent_macro is None:
                    door.debug(macro_exp.traceback)
                    msg = ("An error occurred while running {}:\n"
                           "{!r}").format(macro_obj.getName(), macro_exp)
                    door.error(msg)
                    msg = "Hint: in Spock execute `www`to get more details"
                    door.info(msg)
            self._popMacro()
            raise macro_exp
        self.debug("[ END ] runMacro %s" % desc)
        self._popMacro()

        # decide whether to preserve the macro data
        env_var_name = 'PreserveMacroData'
        try:
            preserve_macro_data = macro_obj.getEnv(env_var_name)
        except UnknownEnv:
            preserve_macro_data = True
        if not preserve_macro_data:
            self.debug('Macro data will not be preserved. ' +
                       'Set "%s" environment variable ' % env_var_name +
                       'to True in order to change it.')
            self._macro_pointer = None

        log_macro_manager.disable()

        return result

    def _popMacro(self):
        self._macro_stack.pop()
        length = len(self._macro_stack)
        if length > 0:
            self._macro_pointer = self._macro_stack[-1]

    def sendState(self, state):
        return self.door.set_state(state)

    def sendStatus(self, status):
        return self.door.set_status(status)

    def sendResult(self, result):
        return self.door.set_result(result)

    def getLastMacroStatus(self):
        return self._macro_pointer._getMacroStatus()

    def sendMacroStatusFinish(self):
        ms = self.getLastMacroStatus()
        if ms is not None:
            ms['state'] = 'finish'

            self.debug("Sending finish event")
            self.sendMacroStatus((ms,))

    def sendMacroStatusStop(self):
        ms = self.getLastMacroStatus()
        if ms is not None:
            ms['state'] = 'stop'

            self.debug("Sending stop event")
            self.sendMacroStatus((ms,))

    def sendMacroStatusAbort(self):
        ms = self.getLastMacroStatus()
        if ms is not None:
            ms['state'] = 'abort'

            self.debug("Sending abort event")
            self.sendMacroStatus((ms,))

    def sendMacroStatusException(self, exc_info):
        ms = self.getLastMacroStatus()
        if ms is not None:
            ms['state'] = 'exception'
            ms['exc_type'] = str(exc_info[0])
            ms['exc_value'] = str(exc_info[1])
            ms['exc_stack'] = traceback.format_exception(*exc_info)
            self.debug("Sending exception event")
            self.sendMacroStatus((ms,))

    def sendMacroStatusPause(self):
        ms = self.getLastMacroStatus()
        if ms is not None and len(ms) > 0:
            ms['state'] = 'pause'
            self.debug("Sending pause event")
            self.sendMacroStatus((ms,))

    def sendMacroStatusResume(self):
        ms = self.getLastMacroStatus()
        if ms is not None and len(ms) > 0:
            ms['state'] = 'resume'
            self.debug("Sending resume event")
            self.sendMacroStatus((ms,))

    def sendMacroStatus(self, data):
        self._last_macro_status = data
        # data = self._macro_status_codec.encode(('', data))
        return self.door.set_macro_status(data)

    def sendRecordData(self, data, codec=None):
        return self.door.set_record_data(data, codec=codec)

    def reserveObj(self, obj, macro_obj, priority=0):
        if obj is None or macro_obj is None:
            return

        # Fill _reserved_macro_objs
        objs = self._reserved_macro_objs[macro_obj] = \
            self._reserved_macro_objs.get(macro_obj, list())
        if obj not in objs:
            if priority:
                objs.insert(0, obj)
            else:
                objs.append(obj)

        # Fill _reserved_objs
        macros = self._reserved_objs[obj] = self._reserved_objs.get(obj, set())
        macros.add(macro_obj)

        # Tell the object that it is reserved by a new macro
        if hasattr(obj, 'reserve'):
            obj.reserve(macro_obj)

    def returnObjs(self, macro_obj):
        """Free the macro reserved objects"""
        if macro_obj is None:
            return
        # remove eventually stopped objects to not keep reference to them
        self._stopped_macro_objs.pop(macro_obj, None)
        objs = self._reserved_macro_objs.get(macro_obj)
        if objs is None:
            return

        # inside returnObj we change the list so we have to iterate with a copy
        for obj in copy.copy(objs):
            self.returnObj(obj, macro_obj)

    def returnObj(self, obj, macro_obj):
        """Free an object reserved by a macro"""
        if obj is None or macro_obj is None:
            return

        if hasattr(obj, 'unreserve'):
            obj.unreserve()
        objs = self._reserved_macro_objs.get(macro_obj)
        if objs is None:
            return
        objs.remove(obj)
        if len(objs) == 0:
            del self._reserved_macro_objs[macro_obj]

        try:
            macros = self._reserved_objs[obj]
            macros.remove(macro_obj)
            if not len(macros):
                del self._reserved_objs[obj]
        except KeyError:
            self.debug("Unexpected KeyError trying to remove reserved object")



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the class for an
acquisition"""

__all__ = ["get_acq_ctrls", "AcquisitionState", "AcquisitionMap",
           "PoolCTAcquisition", "Pool0DAcquisition", "PoolIORAcquisition",
           "PoolAcquisitionHardware", "PoolAcquisitionSoftware",
           "PoolAcquisitionSoftwareStart"]

__docformat__ = 'restructuredtext'

import time
import weakref
import datetime
import traceback
import functools
import threading

from taurus.core.util.log import DebugIt
from taurus.core.util.enumeration import Enumeration

from sardana import AttrQuality, SardanaValue, State, ElementType, \
    TYPE_TIMERABLE_ELEMENTS

from sardana.sardanathreadpool import get_thread_pool
from sardana.pool import AcqSynch, AcqMode
from sardana.pool.poolaction import ActionContext, PoolAction, \
    OperationContext
from sardana.pool.poolsynchronization import PoolSynchronization

#: enumeration representing possible motion states
AcquisitionState = Enumeration("AcquisitionState", (
    "Stopped",
    #    "StoppedOnError",
    #    "StoppedOnAbort",
    "Acquiring",
    "Invalid"))

AS = AcquisitionState
AcquiringStates = AS.Acquiring,
StoppedStates = AS.Stopped,  # MS.StoppedOnError, MS.StoppedOnAbort

AcquisitionMap = {
    # AS.Stopped           : State.On,
    AS.Acquiring: State.Moving,
    AS.Invalid: State.Invalid,
}


def is_value_error(value):
    if isinstance(value, SardanaValue) and value.error:
        return True
    return False


def get_acq_ctrls(ctrls):
    """Converts configuration controllers into acquisition controllers.

    Takes care about converting their internals as well.

    :param ctrls: sequence of configuration controllers objects
    :type ctrls: sardana.pool.poolmeasurementgroup.ControllerConfiguration
    :return: sequence of acquisition controllers
    :rtype: :class:`~sardana.pool.poolacquisition.AcqController`

    .. note::
        The get_acq_ctrls function has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """
    action_ctrls = []
    for ctrl in ctrls:
        action_ctrl = AcqController(ctrl)
        action_ctrls.append(action_ctrl)
    return action_ctrls


def get_timerable_ctrls(ctrls, acq_mode):
    """Converts timerable configuration controllers into acquisition
    controllers.

    Take care about converting their internals as well.
    Take care about assigning master according to acq_mode.

    :param ctrls: sequence of configuration controllers objects
    :type ctrls: sardana.pool.poolmeasurementgroup.ControllerConfiguration
    :param acq_mode: acquisition mode (timer/monitor)
    :type acq_mode: :class:`sardana.pool.AcqMode`
    :return: sequence of acquisition controllers
    :rtype: :class:`~sardana.pool.poolacquisition.AcqController`

    .. note::
        The get_timerable_ctrls function has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """
    action_ctrls = []
    for ctrl in ctrls:
        attrs = {}
        if acq_mode is not None:
            master = None
            if acq_mode is AcqMode.Timer:
                master = ctrl.timer
            elif acq_mode is AcqMode.Monitor:
                master = ctrl.monitor
            attrs = {'master': master}
        action_ctrl = AcqController(ctrl, attrs)
        action_ctrls.append(action_ctrl)
    return action_ctrls


def get_timerable_items(ctrls, master, acq_mode=AcqMode.Timer):
    """Converts timerable configuration items into acquisition items.

    The timerable items are controllers and master. Convert these into
    the corresponding acquisition items.

    Take care about converting their internals as well.
    Take care about assigning master according to acq_mode.

    :param ctrls: sequence of configuration controllers objects
    :type ctrls: :obj:list<:class:`~sardana.pool.poolmeasurementgroup.ControllerConfiguration`>  # noqa
    :param master: master configuration object
    :type master: :class:`~sardana.pool.poolmeasurementgroup.ChannelConfiguration`  # noqa
    :param acq_mode: acquisition mode (timer/monitor)
    :type acq_mode: :class:`sardana.pool.AcqMode`
    :return: sequence of acquisition controllers
    :rtype: :class:`~sardana.pool.poolacquisition.AcqController`

    .. note::
        The get_timerable_ctrls function has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """
    ctrls = get_timerable_ctrls(ctrls, acq_mode)
    # Search master AcqConfigurationItem obj
    for ctrl in ctrls:
        for channel in ctrl.get_channels():
            if channel.configuration == master:
                master = channel
                break
    return ctrls, master


class ActionArgs(object):

    def __init__(self, args, kwargs=None):
        self.args = args
        if kwargs is None:
            kwargs = {}
        self.kwargs = kwargs


class AcqConfigurationItem(object):
    """Wrapper for configuration item that will be used in an action.

    .. note::
        The AcqConfigurationItem function has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, configuration, attrs=None):
        """Constructs action item from a configuration item.

        Eventually it can be enriched with attrs.

        :param configuration: item configuration object
        :type configuration:
            :class:`sardana.pool.poolmeasurementgroup.ConfigurationItem`
        :param attrs: extra attributes to be inserted
        :type attrs: dict
        """
        self._configuration = weakref.ref(configuration)
        self.enabled = True

        if attrs is not None:
            self.__dict__.update(attrs)

    def __getattr__(self, item):
        return getattr(self.configuration, item)

    def get_configuration(self):
        """Returns the element associated with this item"""
        return self._configuration()

    def set_configuration(self, configuration):
        """Sets the element for this item"""
        self._configuration = weakref.ref(configuration)

    configuration = property(get_configuration)


class AcqController(AcqConfigurationItem):
    """Wrapper for controller configuration that will be used in an action.

    .. note::
        The AcqController function has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, configuration, attrs=None):
        """Constructs action controller from a configuration controller.

        Eventually it can be enriched with attrs.

        :param configuration: controller configuration object
        :type configuration:
            :class:`sardana.pool.poolmeasurementgroup.ControllerConfiguration`
        :param attrs: extra attributes to be inserted
        :type attrs: dict
        """
        master = None
        if attrs is not None:
            master = attrs.get('master')
        self._channels = []
        self._channels_enabled = []
        self._channels_disabled = []
        ch_attrs = {'controller': self}
        for conf_channel in configuration.get_channels():
            action_channel = AcqConfigurationItem(conf_channel, ch_attrs)
            self._channels.append(action_channel)
            if conf_channel in configuration.get_channels(enabled=True):
                self._channels_enabled.append(action_channel)
            if conf_channel in configuration.get_channels(enabled=False):
                self._channels_disabled.append(action_channel)
            if master is None:
                continue
            if master == conf_channel:
                attrs['master'] = action_channel
                master = None
        AcqConfigurationItem.__init__(self, configuration, attrs)

    def get_channels(self, enabled=None):
        if enabled is None:
            return list(self._channels)
        elif enabled:
            return list(self._channels_enabled)
        else:
            return list(self._channels_disabled)


class AcquisitionBaseContext(OperationContext):

    def exit(self):
        pool_action = self._pool_action
        pool_action._reset_ctrl_dicts()
        return OperationContext.exit(self)


class PoolAcquisition(PoolAction):
    """Acquisition action which is internally composed for sub-actions.

    Handle acquisition of experimental channels of the following types:
    * timerable (C/T, 1D and 2D) synchronized by software or hardware
    trigger/gate/start
    * 0D

    Synchronized by T/G elements or sofware synchronizer.
    """

    def __init__(self, main_element, name="Acquisition"):
        PoolAction.__init__(self, main_element, name)
        zerodname = name + ".0DAcquisition"
        hwname = name + ".HardwareAcquisition"
        swname = name + ".SoftwareAcquisition"
        sw_start_name = name + ".SoftwareStartAcquisition"
        synchname = name + ".Synchronization"

        self._sw_acq_args = None
        self._sw_start_acq_args = None
        self._0d_acq_args = None
        self._hw_acq_args = None
        self._synch_args = None
        self._sw_acq = PoolAcquisitionSoftware(main_element, name=swname)
        self._sw_start_acq = PoolAcquisitionSoftwareStart(
            main_element, name=sw_start_name)
        self._0d_acq = Pool0DAcquisition(main_element, name=zerodname)
        self._hw_acq = PoolAcquisitionHardware(main_element, name=hwname)
        self._synch = PoolSynchronization(main_element, name=synchname)
        self._handled_first_active = False

    def event_received(self, *args, **kwargs):
        """Callback executed on event of software synchronizer.

        Reacts on start, active, passive or end type of events
        """
        timestamp = time.time()
        _, type_, index = args
        name = type_.name
        if name == "state":
            return
        t_fmt = '%Y-%m-%d %H:%M:%S.%f'
        t_str = datetime.datetime.fromtimestamp(timestamp).strftime(t_fmt)
        msg = '%s event with id: %d received at: %s' % (name, index, t_str)
        self.debug(msg)
        if name == "start":
            if self._sw_start_acq_args is not None:
                self._sw_start_acq._wait()
                self._sw_start_acq._set_busy()
                self.debug('Executing software start acquisition.')
                self._sw_start_acq._started = True
                get_thread_pool().add(self._sw_start_acq.run,
                                      self._sw_start_acq._set_ready,
                                      *self._sw_start_acq_args.args,
                                      **self._sw_start_acq_args.kwargs)
        elif name == "active":
            # this code is not thread safe, but for the moment we assume that
            # only one EventGenerator will work at the same time
            if self._handled_first_active:
                timeout = 0
            else:
                timeout = None
                self._handled_first_active = True
            if self._sw_acq_args is not None:
                if not self._sw_acq._wait(timeout):
                    msg = ('Skipping trigger: software acquisition is still'
                           ' in progress.')
                    self.debug(msg)
                    return
                else:
                    self._sw_acq._set_busy()
                    self.debug('Executing software acquisition.')
                    self._sw_acq_args.kwargs.update({'index': index})
                    self._sw_acq._started = True
                    get_thread_pool().add(self._sw_acq.run,
                                          self._sw_acq._set_ready,
                                          *self._sw_acq_args.args,
                                          **self._sw_acq_args.kwargs)
            if self._0d_acq_args is not None:
                if not self._0d_acq._wait(timeout):
                    msg = ('Skipping trigger: ZeroD acquisition is still in'
                           ' progress.')
                    self.debug(msg)
                    return
                else:
                    self._0d_acq._set_busy()
                    self.debug('Executing ZeroD acquisition.')
                    self._0d_acq_args.kwargs.update({'index': index})
                    self._0d_acq._started = True
                    self._0d_acq._stopped = False
                    self._0d_acq._aborted = False
                    get_thread_pool().add(self._0d_acq.run,
                                          self._0d_acq._set_ready,
                                          *self._0d_acq_args.args,
                                          **self._0d_acq_args.kwargs)
        elif name == "passive":
            # TODO: _0d_acq_args comparison may not be necessary
            if (self._0d_acq_args is not None
                    and not self._0d_acq._is_ready()):
                self.debug('Stopping ZeroD acquisition.')
                self._0d_acq.stop_action()

    def prepare(self, config, acq_mode, value, synch_description=None,
                moveable=None, sw_synch_initial_domain=None,
                nb_starts=1, **kwargs):
        """Prepare measurement process.

        Organize sub-action arguments and loads configuration parameters to
        the hardware controllers.
        """
        self._sw_acq_args = None
        self._sw_start_acq_args = None
        self._0d_acq_args = None
        self._hw_acq_args = None
        self._synch_args = None
        self._handled_first_active = False
        ctrls_hw = []
        ctrls_sw = []
        ctrls_sw_start = []

        repetitions = synch_description.repetitions
        latency = synch_description.passive_time
        # Prepare controllers synchronized by hardware
        acq_sync_hw = [AcqSynch.HardwareTrigger, AcqSynch.HardwareStart,
                       AcqSynch.HardwareGate]
        ctrls = config.get_timerable_ctrls(acq_synch=acq_sync_hw, enabled=True)
        if len(ctrls) > 0:
            ctrls_hw = get_timerable_ctrls(ctrls, acq_mode)
            hw_args = (ctrls_hw, value, repetitions, latency)
            hw_kwargs = {}
            hw_kwargs.update(kwargs)
            self._hw_acq_args = ActionArgs(hw_args, hw_kwargs)

        # Prepare controllers synchronized by software Trigger and Gate
        acq_sync_sw = [AcqSynch.SoftwareGate, AcqSynch.SoftwareTrigger]
        ctrls = config.get_timerable_ctrls(acq_synch=acq_sync_sw, enabled=True)
        if len(ctrls) > 0:
            if acq_mode is AcqMode.Timer:
                master = config.get_master_timer_software()
            elif acq_mode is AcqMode.Monitor:
                master = config.get_master_monitor_software()

            ctrls_sw, master_sw = get_timerable_items(ctrls, master, acq_mode)

            sw_args = (ctrls_sw, value, master_sw)
            sw_kwargs = {'synch': True}
            sw_kwargs.update(kwargs)
            self._sw_acq_args = ActionArgs(sw_args, sw_kwargs)

        # Prepare controllers synchronized by software Start
        ctrls = config.get_timerable_ctrls(acq_synch=AcqSynch.SoftwareStart,
                                           enabled=True)
        if len(ctrls) > 0:
            if acq_mode is AcqMode.Timer:
                master = config.get_master_timer_software_start()
            elif acq_mode is AcqMode.Monitor:
                master = config.get_master_monitor_software_start()

            ctrls_sw_start, master_sw_start = get_timerable_items(ctrls,
                                                                  master,
                                                                  acq_mode)
            sw_start_args = (ctrls_sw_start, value, master_sw_start,
                             repetitions, latency)
            sw_start_kwargs = {'synch': True}
            sw_start_kwargs.update(kwargs)
            self._sw_start_acq_args = ActionArgs(sw_start_args,
                                                 sw_start_kwargs)

        # Prepare 0D controllers
        ctrls = config.get_zerod_ctrls(enabled=True)
        if len(ctrls) > 0:
            ctrls_acq_0d = get_acq_ctrls(ctrls)
            zerod_args = (ctrls_acq_0d,)
            zerod_kwargs = {'synch': True}
            zerod_kwargs.update(kwargs)
            self._0d_acq_args = ActionArgs(zerod_args, zerod_kwargs)

        # Prepare synchronizer controllers
        ctrls = config.get_synch_ctrls(enabled=True)
        ctrls_synch = get_acq_ctrls(ctrls)
        synch_args = (ctrls_synch, synch_description)
        synch_kwargs = {'moveable': moveable,
                        'sw_synch_initial_domain': sw_synch_initial_domain}
        synch_kwargs.update(kwargs)
        self._synch_args = ActionArgs(synch_args, synch_kwargs)

        # Load the configuration to the timerable controllers
        # TODO: apply the configuration only if necessary
        # Checking only the "changed" flag is not enough, one needs to check
        # if the controllers were not used with different measurement groups
        # configurations meanwhile (see: sardana-org/sardana#1171) in this
        # case the configuration must be applied even if it was not changed
        # if config.changed:
        ctrls = ctrls_hw + ctrls_sw_start + ctrls_sw

        for ctrl in ctrls:
            pool_ctrl = ctrl.element
            if not pool_ctrl.is_online():
                raise RuntimeError('The controller {0} is '
                                   'offline'.format(pool_ctrl.name))
            pool_ctrl.set_ctrl_par('acquisition_mode', acq_mode)
            pool_ctrl.operator = self.main_element
            pool_ctrl.set_ctrl_par('timer', ctrl.timer.axis)
            pool_ctrl.set_ctrl_par('monitor', ctrl.monitor.axis)
            synch = config.get_acq_synch_by_controller(pool_ctrl)
            pool_ctrl.set_ctrl_par('synchronization', synch)

            if ctrl.is_referable():
                for channel in ctrl.get_channels():
                    value_ref_enabled = channel.value_ref_enabled
                    pool_ctrl.set_axis_par(channel.axis,
                                           "value_ref_enabled",
                                           value_ref_enabled)
                    if value_ref_enabled:
                        pool_ctrl.set_axis_par(channel.axis,
                                               "value_ref_pattern",
                                               channel.value_ref_pattern)

        config.changed = False

        # Call hardware and software start controllers prepare method
        ctrls = ctrls_hw + ctrls_sw_start
        self._prepare_ctrls(ctrls, value, repetitions, latency,
                            nb_starts)

        # Call software controllers prepare method
        nb_starts = nb_starts * repetitions
        repetitions = 1
        self._prepare_ctrls(ctrls_sw, value, repetitions, latency,
                            nb_starts)

    @staticmethod
    def _prepare_ctrls(ctrls, value, repetitions, latency, nb_starts):
        for ctrl in ctrls:
            axis = ctrl.master.axis
            pool_ctrl = ctrl.element
            pool_ctrl.ctrl.PrepareOne(axis, value, repetitions, latency,
                                      nb_starts)

    def is_running(self):
        """Checks if acquisition is running.

        Acquisition is runnin if any of its sub-actions is running.
        """
        return self._sw_start_acq.is_running()\
            or self._0d_acq.is_running()\
            or self._sw_acq.is_running()\
            or self._hw_acq.is_running()\
            or self._synch.is_running()

    def run(self, *args, **kwargs):
        """Runs acquisition according to previous preparation."""
        for elem in self.get_elements():
            elem.put_state(None)
            # TODO: temporarily clear value buffers at the beginning of the
            # acquisition instead of doing it in the finish hook of each
            # acquisition sub-actions. See extensive explanation in the
            # constructor of PoolAcquisitionBase.
            try:
                elem.clear_value_buffer()
            except AttributeError:
                continue
            try:
                elem.clear_value_ref_buffer()
            except AttributeError:
                continue
            # clean also the pseudo counters, even the ones that do not
            # participate directly in the acquisition
            for pseudo_elem in elem.get_pseudo_elements():
                pseudo_elem.clear_value_buffer()

        if self._hw_acq_args is not None:
            self._hw_acq._wait()
            self._hw_acq._set_busy()
            self._hw_acq.run(*self._hw_acq_args.args,
                             **self._hw_acq_args.kwargs,
                             cb=self._hw_acq._set_ready)

        if self._sw_acq_args is not None\
                or self._sw_start_acq_args is not None\
                or self._0d_acq_args is not None:
            self._synch.add_listener(self)

        if self._synch_args is not None:
            self._synch._wait()
            self._synch._set_busy()
            self._synch.run(*self._synch_args.args,
                            **self._synch_args.kwargs,
                            cb=self._synch._set_ready)

    def _get_action_for_element(self, element):
        elem_type = element.get_type()
        if elem_type in TYPE_TIMERABLE_ELEMENTS:
            config = self.main_element.configuration
            try:
                acq_synch = config.get_acq_synch_by_channel(element)
            # when configuration was not yet set and one sets the
            # measurement group's integration time (this may happen on Tango
            # device initialization when memorized attributes are set we
            # fallback to software acquisition
            except KeyError:
                acq_synch = AcqSynch.SoftwareTrigger
            if acq_synch in (AcqSynch.SoftwareTrigger,
                             AcqSynch.SoftwareGate):
                return self._sw_acq
            elif acq_synch == AcqSynch.SoftwareStart:
                return self._sw_start_acq
            elif acq_synch in (AcqSynch.HardwareTrigger,
                               AcqSynch.HardwareGate,
                               AcqSynch.HardwareStart):
                return self._hw_acq
        elif elem_type == ElementType.ZeroDExpChannel:
            return self._0d_acq
        elif elem_type == ElementType.TriggerGate:
            return self._synch
        else:
            raise RuntimeError("Could not determine action for element %s" %
                               element)

    def clear_elements(self):
        """Clears all elements from this action"""

    def add_element(self, element):
        """Adds a new element to this action.

        :param element: the new element to be added
        :type element: sardana.pool.poolelement.PoolElement"""
        action = self._get_action_for_element(element)
        action.add_element(element)

    def remove_element(self, element):
        """Removes an element from this action. If the element is not part of
        this action, a ValueError is raised.

        :param element: the new element to be removed
        :type element: sardana.pool.poolelement.PoolElement

        :raises: ValueError"""
        for action in self._get_acq_for_element(element):
            action.remove_element(element)

    def get_elements(self, copy_of=False):
        """Returns a sequence of all elements involved in this action.

        :param copy_of: If False (default) the internal container of
                        elements is returned. If True, a copy of the
                        internal container is returned instead
        :type copy_of: bool
        :return: a sequence of all elements involved in this action.
        :rtype: seq<sardana.pool.poolelement.PoolElement>"""
        return (self._hw_acq.get_elements() + self._sw_acq.get_elements()
                + self._sw_start_acq.get_elements()
                + self._0d_acq.get_elements() + self._synch.get_elements())

    def get_pool_controller_list(self):
        """Returns a list of all controller elements involved in this action.

        :return: a list of all controller elements involved in this action.
        :rtype: list<sardana.pool.poolelement.PoolController>"""
        return self._pool_ctrl_list

    def get_pool_controllers(self):
        """Returns a dict of all controller elements involved in this action.

        :return: a dict of all controller elements involved in this action.
        :rtype: dict<sardana.pool.poolelement.PoolController,
                seq<sardana.pool.poolelement.PoolElement>>"""
        ret = {}
        ret.update(self._hw_acq.get_pool_controllers())
        ret.update(self._sw_acq.get_pool_controllers())
        ret.update(self._sw_start_acq.get_pool_controllers())
        ret.update(self._0d_acq.get_pool_controllers())
        return ret

    def read_value(self, ret=None, serial=False):
        """Reads value information of all elements involved in this action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement`,
                     :class:~`sardana.sardanavalue.SardanaValue`>"""
        # TODO: this is broken now - fix it
        ret = self._ct_acq.read_value(ret=ret, serial=serial)
        ret.update(self._0d_acq.read_value(ret=ret, serial=serial))
        return ret


class PoolAcquisitionBase(PoolAction):
    """Base class for sub-acquisition.

    .. note::
        The PoolAcquisitionBase class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.

    .. todo: Think of moving the ready/busy mechanism to PoolAction
    """

    def __init__(self, main_element, name):
        PoolAction.__init__(self, main_element, name)
        self._channels = []
        self._index = None
        self._ready = threading.Event()
        self._ready.set()

    def _is_ready(self):
        return self._ready.is_set()

    def _wait(self, timeout=None):
        return self._ready.wait(timeout)

    def _set_ready(self, _=None):
        self._ready.set()

    def _is_busy(self):
        return not self._ready.is_set()

    def _set_busy(self):
        self._ready.clear()


class PoolAcquisitionTimerable(PoolAcquisitionBase):
    """Base class for acquisitions of timerable channels.

     Implements a generic start_action method. action_loop method must be
     implemented by the sub-class.

    .. note::
        The PoolAcquisitionTimerable class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    OperationContextClass = AcquisitionBaseContext

    def __init__(self, main_element, name):
        PoolAcquisitionBase.__init__(self, main_element, name)
        self._nb_states_per_value = None
        self._acq_sleep_time = None
        self._pool_ctrl_dict_loop = None
        self._pool_ctrl_dict_ref = None
        self._pool_ctrl_dict_value = None

        # TODO: for the moment we can not clear value buffers at the end of
        # the acquisition. This is because of the pseudo counters that are
        # based on channels synchronized by hardware and software.
        # These two acquisition actions finish at different moment so the
        # pseudo counter will loose the value buffer of some of its physicals
        # if we clear the buffer at the end.
        # Whenever there will be solution for that, after refactoring of the
        # acquisition actions, uncomment this line
        # self.add_finish_hook(self.clear_value_buffers, True)

    def get_read_value_ref_ctrls(self):
        return self._pool_ctrl_dict_ref

    def read_value_ref(self, ret=None, serial=False):
        """Reads value ref information of all elements involved in this action

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement`,
                     (value object, Exception or None)>"""
        with ActionContext(self):
            return self.raw_read_value_ref(ret=ret, serial=serial)

    def raw_read_value_ref(self, ret=None, serial=False):
        """**Unsafe**. Reads value ref information of all referable elements
        involved in this acquisition

        :param ret: output map parameter that should be filled with value
                    information. If None is given (default), a new map is
                    created an returned
        :type ret: dict
        :param serial: If False (default) perform controller HW value requests
                       in parallel. If True, access is serialized.
        :type serial: bool
        :return: a map containing value information per element
        :rtype: dict<:class:~`sardana.pool.poolelement.PoolElement,
                :class:`sardana.sardanavalue.SardanaValue`>
        """
        if ret is None:
            ret = {}

        read = self._raw_read_value_ref_concurrent
        if serial:
            read = self._raw_read_value_ref_serial

        value_info = self._value_info

        with value_info:
            value_info.init(len(self.get_read_value_ref_ctrls()))
            read(ret)
            value_info.wait()
        return ret

    def _raw_read_value_ref_serial(self, ret):
        """Internal method. Read value ref in a serial mode"""
        for pool_ctrl in self.get_read_value_ref_ctrls():
            self._raw_read_ctrl_value_ref(ret, pool_ctrl)
        return ret

    def _raw_read_value_ref_concurrent(self, ret):
        """Internal method. Read value ref in a concurrent mode"""
        th_pool = get_thread_pool()
        for pool_ctrl in self.get_read_value_ref_ctrls():
            th_pool.add(self._raw_read_ctrl_value_ref, None, ret, pool_ctrl)
        return ret

    def _raw_read_ctrl_value_ref(self, ret, pool_ctrl):
        """Internal method. Read controller value ref information and store
        it in ret parameter"""
        try:
            axes = [elem.axis for elem in self._pool_ctrl_dict_ref[pool_ctrl]]
            value_infos = pool_ctrl.raw_read_axis_value_refs(axes)
            ret.update(value_infos)
        finally:
            self._value_info.finish_one()

    def in_acquisition(self, states):
        """Determines if we are in acquisition or if the acquisition has ended
        based on the current unit trigger modes and states returned by the
        controller(s)

        :param states: a map containing state information as returned by
                       read_state_info
        :type states: dict<PoolElement, State>
        :return: returns True if in acquisition or False otherwise
        :rtype: bool"""
        for elem in states:
            s = states[elem][0][0]
            if self._is_in_action(s):
                return True

    @DebugIt()
    def start_action(self, ctrls, value, master, repetitions, latency,
                     index, acq_sleep_time, nb_states_per_value,
                     **kwargs):
        """
        Prepares everything for acquisition and starts it
        :param ctrls: List of enabled pool acquisition controllers
        :type ctrls: list
        :param value: integration time/monitor counts
        :type value: float/int or seq<float/int>
        :param repetitions: repetitions
        :type repetitions: int
        :param latency:
        :type latency: float
        :param master: master channel is the last one to start
        :type master: Channel
        :param index:
        :type index: int
        :param acq_sleep_time: sleep time between state queries
        :type acq_sleep_time: float
        :param nb_states_per_value: how many state queries between readouts
        :type nb_states_per_value: int
        :param args:
        :param kwargs:
        :return:
        """

        pool = self.pool
        self._aborted = False
        self._stopped = False

        self._index = index

        self._acq_sleep_time = acq_sleep_time
        if self._acq_sleep_time is None:
            self._acq_sleep_time = pool.acq_loop_sleep_time

        self._nb_states_per_value = nb_states_per_value
        if self._nb_states_per_value is None:
            self._nb_states_per_value = pool.acq_loop_states_per_value

        # make sure the controller which has the master channel is the last to
        # be called
        if master is not None:
            ctrls.remove(master.controller)
            ctrls.append(master.controller)

        # controllers that will be read during the action
        self._set_pool_ctrl_dict_loop(ctrls)
        # split controllers to read value and value reference
        self._split_ctrl(ctrls)
        # channels that are acquired (only enabled)
        self._channels = []

        def load(channel, value, repetitions, latency=0):
            axis = channel.axis
            pool_ctrl = channel.controller
            ctrl = pool_ctrl.ctrl
            ctrl.PreLoadAll()
            res = ctrl.PreLoadOne(axis, value, repetitions, latency)
            if not res:
                msg = ("%s.PreLoadOne(%d) returned False" %
                       (pool_ctrl.name, axis))
                raise Exception(msg)
            ctrl.LoadOne(axis, value, repetitions, latency)
            ctrl.LoadAll()

        with ActionContext(self):
            # PreLoadAll, PreLoadOne, LoadOne and LoadAll
            for ctrl in ctrls:
                # TODO find solution for master now sardana only use timer
                load(ctrl.timer, value, repetitions, latency)

            # TODO: remove when the action allows to use tango attributes
            try:
                ctrls.pop('__tango__')
            except Exception:
                pass

            # PreStartAll on all enabled controllers
            for ctrl in ctrls:
                pool_ctrl = ctrl.element
                pool_ctrl.ctrl.PreStartAll()

            # PreStartOne & StartOne on all enabled elements
            for ctrl in ctrls:
                channels = ctrl.get_channels(enabled=True)

                # make sure that the master timer/monitor is started as
                # the last one
                channels.remove(ctrl.master)
                channels.append(ctrl.master)
                for channel in channels:
                    axis = channel.axis
                    pool_ctrl = ctrl.element
                    ret = pool_ctrl.ctrl.PreStartOne(axis, value)
                    if not ret:
                        msg = ("%s.PreStartOne(%d) returns False" %
                               (ctrl.name, axis))
                        raise Exception(msg)
                    try:
                        pool_ctrl = ctrl.element
                        pool_ctrl.ctrl.StartOne(axis, value)
                    except Exception as e:
                        self.debug(e, exc_info=True)
                        channel.set_state(State.Fault, propagate=2)
                        msg = ("%s.StartOne(%d) failed" %
                               (ctrl.name, axis))
                        raise Exception(msg)

                    self._channels.append(channel)

            # set the state of all elements to  and inform their listeners
            for channel in self._channels:
                channel.set_state(State.Moving, propagate=2)

            # StartAll on all enabled controllers
            for ctrl in ctrls:
                try:
                    pool_ctrl = ctrl.element
                    pool_ctrl.ctrl.StartAll()
                except Exception as e:
                    channels = ctrl.get_channels(enabled=True)
                    self.debug(e, exc_info=True)
                    for channel in channels:
                        channel.set_state(State.Fault, propagate=2)
                    msg = ("%s.StartAll() failed" % ctrl.name)
                    raise Exception(msg)

    def _set_pool_ctrl_dict_loop(self, ctrls):
        ctrl_channels = {}
        for ctrl in ctrls:
            pool_channels = []
            pool_ctrl = ctrl.element
            # only CT will be read in the loop, 1D and 2D not
            if ElementType.CTExpChannel not in ctrl.get_ctrl_types():
                continue
            for channel in ctrl.get_channels(enabled=True):
                pool_channels.append(channel.element)
            ctrl_channels[pool_ctrl] = pool_channels
        self._pool_ctrl_dict_loop = ctrl_channels

    def _split_ctrl(self, ctrls):
        ctrl_channels_value = {}
        ctrl_channels_ref = {}
        for ctrl in ctrls:
            if not ctrl.is_referable():
                pool_channels_value = []
                pool_ctrl = ctrl.element
                for channel in ctrl.get_channels(enabled=True):
                    pool_channels_value.append(channel.element)
                ctrl_channels_value[pool_ctrl] = pool_channels_value
            else:
                pool_channels_value = []
                pool_channels_ref = []
                pool_ctrl = ctrl.element
                for channel in ctrl.get_channels(enabled=True):
                    if channel.value_ref_enabled:
                        pool_channels_ref.append(channel.element)
                        if channel.has_pseudo_elements():
                            pool_channels_value.append(channel.element)
                    else:
                        pool_channels_value.append(channel.element)
                if len(pool_channels_value) > 0:
                    ctrl_channels_value[pool_ctrl] = pool_channels_value
                if len(pool_channels_ref) > 0:
                    ctrl_channels_ref[pool_ctrl] = pool_channels_ref

        self._pool_ctrl_dict_value = ctrl_channels_value
        self._pool_ctrl_dict_ref = ctrl_channels_ref

    def _reset_ctrl_dicts(self):
        self._pool_ctrl_dict_loop = None
        self._pool_ctrl_dict_value = None
        self._pool_ctrl_dict_ref = None

    def clear_value_buffers(self):
        for channel in self._channels:
            channel.clear_value_buffer()


class PoolAcquisitionHardware(PoolAcquisitionTimerable):
    """Acquisition action for controllers synchronized by hardware

    .. note::
        The PoolAcquisitionHardware class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.

    .. todo:: Try to move the action loop logic to base class it is
    basically the same as in PoolAcquisitionSoftwareStart.
    """

    def __init__(self, main_element, name="AcquisitionHardware"):
        PoolAcquisitionTimerable.__init__(self, main_element, name)

    def start_action(self, ctrls, value, repetitions, latency,
                     acq_sleep_time=None, nb_states_per_value=None,
                     **kwargs):
        PoolAcquisitionTimerable.start_action(self, ctrls, value, None,
                                         repetitions, latency, None,
                                         acq_sleep_time, nb_states_per_value,
                                         **kwargs)

    def get_read_value_ctrls(self):
        return self._pool_ctrl_dict_value

    @DebugIt()
    def action_loop(self):
        i = 0

        states, values, value_refs = {}, {}, {}
        for channel in self._channels:
            element = channel.element
            states[element] = None

        nap = self._acq_sleep_time
        nb_states_per_value = self._nb_states_per_value

        while True:
            self.read_state_info(ret=states)
            if not self.in_acquisition(states):
                break

            # read value every n times
            if not i % nb_states_per_value:
                self.read_value(ret=values)
                for acquirable, value in list(values.items()):
                    if is_value_error(value):
                        self.error("Loop read value error for %s" %
                                   acquirable.name)
                        msg = "Details: " + "".join(
                            traceback.format_exception(*value.exc_info))
                        self.debug(msg)
                        acquirable.put_value(value)
                    else:
                        acquirable.extend_value_buffer(value)

            time.sleep(nap)
            i += 1

        with ActionContext(self):
            self.raw_read_value(ret=values)
            self.raw_read_value_ref(ret=value_refs)

        for acquirable, state_info in list(states.items()):
            if acquirable in values:
                value = values[acquirable]
                if is_value_error(value):
                    self.error("Loop final read value error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value.exc_info))
                    self.debug(msg)
                    acquirable.put_value(value)
                else:
                    acquirable.extend_value_buffer(value, propagate=2)
            if acquirable in value_refs:
                value_ref = value_refs[acquirable]
                if is_value_error(value_ref):
                    self.error("Loop final read value ref error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value_ref.exc_info))
                    self.debug(msg)
                acquirable.extend_value_ref_buffer(value_ref, propagate=2)
            state_info = acquirable._from_ctrl_state_info(state_info)
            set_state_info = functools.partial(acquirable.set_state_info,
                                               state_info,
                                               propagate=2,
                                               safe=True)
            self.add_finish_hook(set_state_info, False)


class PoolAcquisitionSoftware(PoolAcquisitionTimerable):
    """Acquisition action for controllers synchronized by software

    .. note::
        The PoolAcquisitionSoftware class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, main_element, name="AcquisitionSoftware", slaves=None):
        PoolAcquisitionTimerable.__init__(self, main_element, name)

        if slaves is None:
            slaves = ()
        self._slaves = slaves

    def get_read_value_ctrls(self):
        # technical debt in order to work both in case of meas group and
        # single channel
        if self._pool_ctrl_dict_value is not None:
            return self._pool_ctrl_dict_value
        else:
            return self._pool_ctrl_dict

    def get_read_value_ref_ctrls(self):
        # technical debt in order to work both in case of meas group and
        # single channel
        if self._pool_ctrl_dict_ref is not None:
            return self._pool_ctrl_dict_ref
        else:
            return self._pool_ctrl_dict

    def get_read_value_loop_ctrls(self):
        return self._pool_ctrl_dict_loop

    def start_action(self, ctrls, value, master, index, acq_sleep_time=None,
                     nb_states_per_value=None, **kwargs):
        PoolAcquisitionTimerable.start_action(self, ctrls, value, master, 1, 0,
                                         index, acq_sleep_time,
                                         nb_states_per_value, **kwargs)

    @DebugIt()
    def action_loop(self):
        states, values, value_refs = {}, {}, {}
        for channel in self._channels:
            element = channel.element
            states[element] = None

        nap = self._acq_sleep_time
        nb_states_per_value = self._nb_states_per_value

        i = 0
        while True:
            self.read_state_info(ret=states)
            if not self.in_acquisition(states):
                break

            # read value every n times
            if not i % nb_states_per_value:
                self.read_value_loop(ret=values)
                for acquirable, value in list(values.items()):
                    acquirable.put_value(value, quality=AttrQuality.Changing)

            time.sleep(nap)
            i += 1

        for slave in self._slaves:
            try:
                slave.stop_action()
            except Exception:
                self.warning("Unable to stop slave acquisition %s",
                             slave.getLogName())
                self.debug("Details", exc_info=1)

        with ActionContext(self):
            self.raw_read_value(ret=values)
            self.raw_read_value_ref(ret=value_refs)

        for acquirable, state_info in list(states.items()):
            if acquirable in values:
                value = values[acquirable]
                if is_value_error(value):
                    self.error("Loop final read value error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value.exc_info))
                    self.debug(msg)
                acquirable.get_value_attribute().set_quality(
                    AttrQuality.Valid)
                acquirable.append_value_buffer(value, self._index,
                                               propagate=2)
            if acquirable in value_refs:
                value_ref = value_refs[acquirable]
                if is_value_error(value_ref):
                    self.error("Loop final read value ref error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value_ref.exc_info))
                    self.debug(msg)
                acquirable.append_value_ref_buffer(value_ref, self._index)
            state_info = acquirable._from_ctrl_state_info(state_info)
            set_state_info = functools.partial(acquirable.set_state_info,
                                               state_info,
                                               propagate=2,
                                               safe=True)
            self.add_finish_hook(set_state_info, False)


class PoolAcquisitionSoftwareStart(PoolAcquisitionTimerable):
    """Acquisition action for controllers synchronized by software start

    .. note::
        The PoolAcquisitionSoftwareStart class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.

    .. todo:: Try to move the action loop logic to base class it is
    basically the same as in PoolAcquisitionHardware.
    """

    def __init__(self, main_element, name="AcquisitionSoftwareStart"):
        PoolAcquisitionTimerable.__init__(self, main_element, name)

    def get_read_value_ctrls(self):
        # technical debt in order to work both in case of meas group and
        # single channel
        return self._pool_ctrl_dict_value

    def start_action(self, ctrls, value, master, repetitions, latency,
                     acq_sleep_time=None, nb_states_per_value=None,
                     **kwargs):
        PoolAcquisitionTimerable.start_action(self, ctrls, value, master,
                                         repetitions, latency, None,
                                         acq_sleep_time, nb_states_per_value,
                                         **kwargs)

    @DebugIt()
    def action_loop(self):
        i = 0

        states, values, value_refs = {}, {}, {}
        for channel in self._channels:
            element = channel.element
            states[element] = None

        nap = self._acq_sleep_time
        nb_states_per_value = self._nb_states_per_value

        while True:
            self.read_state_info(ret=states)
            if not self.in_acquisition(states):
                break

            # read value every n times
            if not i % nb_states_per_value:
                self.read_value(ret=values)
                for acquirable, value in list(values.items()):
                    if is_value_error(value):
                        self.error("Loop read value error for %s" %
                                   acquirable.name)
                        msg = "Details: " + "".join(
                            traceback.format_exception(*value.exc_info))
                        self.debug(msg)
                        acquirable.put_value(value)
                    else:
                        acquirable.extend_value_buffer(value)
                self.read_value_ref(ret=value_refs)
                for acquirable, value_ref in list(value_refs.items()):
                    if is_value_error(value_ref):
                        self.error("Loop read value ref error for %s" %
                                   acquirable.name)
                        msg = "Details: " + "".join(
                            traceback.format_exception(*value.exc_info))
                        self.debug(msg)
                        acquirable.put_value_ref(value)
                    else:
                        acquirable.extend_value_ref_buffer(value_ref)
            time.sleep(nap)
            i += 1

        with ActionContext(self):
            self.raw_read_value(ret=values)
            self.raw_read_value_ref(ret=value_refs)

        for acquirable, state_info in list(states.items()):
            if acquirable in values:
                value = values[acquirable]
                if is_value_error(value):
                    self.error("Loop final read value error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value.exc_info))
                    self.debug(msg)
                    acquirable.put_value(value)
                else:
                    acquirable.extend_value_buffer(value, propagate=2)
            if acquirable in value_refs:
                value_ref = value_refs[acquirable]
                if is_value_error(value_ref):
                    self.error("Loop final read value ref error for: %s" %
                               acquirable.name)
                    msg = "Details: " + "".join(
                        traceback.format_exception(*value_ref.exc_info))
                    self.debug(msg)
                    acquirable.put_value_ref(value_ref)
                else:
                    acquirable.extend_value_ref_buffer(value_ref, propagate=2)
            state_info = acquirable._from_ctrl_state_info(state_info)
            set_state_info = functools.partial(acquirable.set_state_info,
                                               state_info,
                                               propagate=2,
                                               safe=True)
            self.add_finish_hook(set_state_info, False)


class PoolCTAcquisition(PoolAcquisitionTimerable):
    """..todo:: remove it, still used by pseudo counter"""

    def __init__(self, main_element, name="CTAcquisition", slaves=None):
        self._channels = None

        if slaves is None:
            slaves = ()
        self._slaves = slaves

        PoolAcquisitionTimerable.__init__(self, main_element, name)

    def get_read_value_loop_ctrls(self):
        return self._pool_ctrl_dict_loop

    def in_acquisition(self, states):
        """Determines if we are in acquisition or if the acquisition has ended
        based on the current unit trigger modes and states returned by the
        controller(s)

        :param states: a map containing state information as returned by
                       read_state_info
        :type states: dict<PoolElement, State>
        :return: returns True if in acquisition or False otherwise
        :rtype: bool"""
        for elem in states:
            s = states[elem][0][0]
            if self._is_in_action(s):
                return True

    @DebugIt()
    def action_loop(self):
        i = 0

        states, values = {}, {}
        for element in self._channels:
            states[element] = None
            # values[element] = None

        nap = self._acq_sleep_time
        nb_states_per_value = self._nb_states_per_value

        # read values to send a first event when starting to acquire
        with ActionContext(self):
            self.raw_read_value_loop(ret=values)
            for acquirable, value in list(values.items()):
                acquirable.put_value(value, propagate=2)

        while True:
            self.read_state_info(ret=states)
            if not self.in_acquisition(states):
                break

            # read value every n times
            if not i % nb_states_per_value:
                self.read_value_loop(ret=values)
                for acquirable, value in list(values.items()):
                    acquirable.put_value(value)

            time.sleep(nap)
            i += 1

        for slave in self._slaves:
            try:
                slave.stop_action()
            except Exception:
                self.warning("Unable to stop slave acquisition %s",
                             slave.getLogName())
                self.debug("Details", exc_info=1)

        with ActionContext(self):
            self.raw_read_state_info(ret=states)
            self.raw_read_value_loop(ret=values)

        for acquirable, state_info in list(states.items()):
            # first update the element state so that value calculation
            # that is done after takes the updated state into account
            acquirable.set_state_info(state_info, propagate=0)
            if acquirable in values:
                value = values[acquirable]
                acquirable.put_value(value, propagate=2)
            state_info = acquirable._from_ctrl_state_info(state_info)
            set_state_info = functools.partial(acquirable.set_state_info,
                                               state_info,
                                               propagate=2,
                                               safe=True)
            self.add_finish_hook(set_state_info, False)


class Pool0DAcquisition(PoolAcquisitionBase):

    def __init__(self, main_element, name="0DAcquisition"):
        PoolAcquisitionBase.__init__(self, main_element, name)

    def start_action(self, conf_ctrls, index, acq_sleep_time=None,
                     nb_states_per_value=None, **kwargs):
        """Prepares everything for acquisition and starts it.

           :param: config"""

        pool = self.pool
        # TODO: rollback this change when a proper synchronization between
        # acquisition actions will be develop.
        # Now the meta acquisition action is resettung them to 0.
        # self._aborted = False
        # self._stopped = False

        self._index = index
        self._acq_sleep_time = acq_sleep_time
        if self._acq_sleep_time is None:
            self._acq_sleep_time = pool.acq_loop_sleep_time

        self._nb_states_per_value = nb_states_per_value
        if self._nb_states_per_value is None:
            self._nb_states_per_value = pool.acq_loop_states_per_value

        # channels that are acquired (only enabled)
        self._channels = []

        with ActionContext(self):
            # set the state of all elements to  and inform their listeners

            for conf_ctrl in conf_ctrls:
                for conf_channel in conf_ctrl.get_channels(enabled=True):
                    conf_channel.clear_buffer()
                    conf_channel.set_state(State.Moving, propagate=2)
                    self._channels.append(conf_channel)

    def in_acquisition(self, states):
        """Determines if we are in acquisition or if the acquisition has ended
        based on the current unit trigger modes and states returned by the
        controller(s)

        :param states: a map containing state information as returned by
                       read_state_info
        :type states: dict<PoolElement, State>
        :return: returns True if in acquisition or False otherwise
        :rtype: bool"""
        for state in states:
            s = states[state][0]
            if self._is_in_action(s):
                return True

    def action_loop(self):
        states, values = {}, {}
        for conf_channel in self._channels:
            element = conf_channel.element
            states[element] = None
            values[element] = None

        nap = self._acq_sleep_time
        while True:
            self.read_value(ret=values)
            for acquirable, value in list(values.items()):
                acquirable.put_current_value(value, propagate=0)
            if self._stopped or self._aborted:
                break
            time.sleep(nap)

        for element in self._channels:
            value = element.accumulated_value.value_obj
            element.append_value_buffer(value, self._index, propagate=2)

        with ActionContext(self):
            self.raw_read_state_info(ret=states)

        for acquirable, state_info in list(states.items()):
            state_info = acquirable._from_ctrl_state_info(state_info)
            set_state_info = functools.partial(acquirable.set_state_info,
                                               state_info,
                                               propagate=2,
                                               safe=True)
            self.add_finish_hook(set_state_info, False)

    def stop_action(self, *args, **kwargs):
        """Stop procedure for this action."""
        self._stopped = True

    def abort_action(self, *args, **kwargs):
        """Aborts procedure for this action"""
        self._aborted = True


class PoolIORAcquisition(PoolAction):

    def __init__(self, pool, name="IORAcquisition"):
        self._channels = None
        PoolAction.__init__(self, pool, name)

    def start_action(self, *args, **kwargs):
        pass

    def in_acquisition(self, states):
        return True
        pass

    @DebugIt()
    def action_loop(self):
        i = 0

        states, values = {}, {}
        for element in self._channels:
            states[element] = None
            values[element] = None

        # read values to send a first event when starting to acquire
        self.read_value(ret=values)
        for acquirable, value in list(values.items()):
            acquirable.put_value(value, propagate=2)

        while True:
            self.read_state_info(ret=states)

            if not self.in_acquisition(states):
                break

            # read value every n times
            if not i % 5:
                self.read_value(ret=values)
                for acquirable, value in list(values.items()):
                    acquirable.put_value(value)

            i += 1
            time.sleep(0.01)

        self.read_state_info(ret=states)

        # first update the element state so that value calculation
        # that is done after takes the updated state into account
        for acquirable, state_info in list(states.items()):
            acquirable.set_state_info(state_info, propagate=0)

        # Do NOT send events before we exit the OperationContext, otherwise
        # we may be asked to start another action before we leave the context
        # of the current action. Instead, send the events in the finish hook
        # which is executed outside the OperationContext

        def finish_hook(*args, **kwargs):
            # read values and propagate the change to all listeners
            self.read_value(ret=values)
            for acquirable, value in list(values.items()):
                acquirable.put_value(value, propagate=2)

            # finally set the state and propagate to all listeners
            for acquirable, state_info in list(states.items()):
                acquirable.set_state_info(state_info, propagate=2)

        self.set_finish_hook(finish_hook)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the macro server scan data output recorder module"""

__all__ = ["NXxas_FileRecorder"]

__docformat__ = 'restructuredtext'

from sardana.macroserver.scan.recorder import BaseNEXUS_FileRecorder


class NXxas_FileRecorder(BaseNEXUS_FileRecorder):
    """saves data to a nexus file that follows the NXsas application definition

        """

    def __init__(self, filename=None, macro=None, overwrite=False, **pars):
        BaseNEXUS_FileRecorder.__init__(
            self, filename=filename, macro=macro, overwrite=overwrite, **pars)

    def _startRecordList(self, recordlist):
        nxs = self.nxs
        if self.filename is None:
            return

        # get the recordlist environment
        self.currentlist = recordlist
        env = self.currentlist.getEnviron()

        # adapt the datadesc to the NeXus requirements
        self.datadesc = []
        for dd in env['datadesc']:
            dd = dd.clone()
            dd.label = self.sanitizeName(dd.label)
            if dd.dtype == 'bool':
                dd.dtype = 'int8'
                self.debug('%s will be stored with type=%s', dd.name, dd.dtype)
            if dd.dtype in self.supported_dtypes:
                self.datadesc.append(dd)
            else:
                self.warning(
                    '%s will not be stored. Reason: type %s not supported', dd.name, dd.dtype)

        serialno = env["serialno"]
        nxfilemode = self.getFormat()
        if not self.overwrite and os.path.exists(self.filename):
            nxfilemode = 'rw'

        self.debug("starting new recording %d on file %s",
                   serialno, self.filename)

        # create an nxentry and write it to file
        self.nxentry = nxs.NXentry(name="entry%d" % serialno)
        self.nxentry.save(self.filename, format=nxfilemode)

        # add fields to nxentry
        import sardana.release
        program_name = "%s (%s)" % (
            sardana.release.name, self.__class__.__name__)
        self.nxentry.insert(nxs.NXfield(name='start_time',
                                        value=env['starttime'].isoformat()))
        self.nxentry.insert(nxs.NXfield(name='title', value=env['title']))
        self.nxentry.insert(nxs.NXfield(name='definition', value='NXxas'))
        self.nxentry.insert(nxs.NXfield(
            name='epoch', value=time.mktime(env['starttime'].timetuple())))
        self.nxentry.insert(nxs.NXfield(name='program_name', value=program_name, attrs={
                            'version': sardana.release.version}))
        self.nxentry.insert(nxs.NXfield(
            name='entry_identifier', value=env['serialno']))

        # add the "measurement" group (a NXcollection containing all counters
        # from the mntgrp for convenience)
        measurement = nxs.NXcollection(name='measurement')
        self.ddfieldsDict = {}
        for dd in self.datadesc:
            field = NXfield_comp(name=dd.label,
                                 dtype=dd.dtype,
                                 shape=[nxs.UNLIMITED] + list(dd.shape),
                                 nxslab_dims=[1] + list(dd.shape)
                                 )
            if hasattr(dd, 'data_units'):
                field.attrs['units'] = dd.data_units
            measurement.insert(field)
            # create a dict of fields in the datadesc for easier access later
            # on
            self.ddfieldsDict[dd.label] = field

        self.nxentry.insert(measurement)

        # user group
        nxuser = nxs.NXuser()
        self.nxentry.insert(nxuser)
        nxuser['name'] = env['user']

        # sample group
        nxsample = nxs.NXsample()
        self.nxentry.insert(nxsample)
        nxsample['name'] = env['SampleInfo'].get('name', 'Unknown')

        # monitor group
        scan_acq_time = env.get('integ_time')
        scan_monitor_mode = scan_acq_time > 1 and 'timer' or 'monitor'
        nxmonitor = nxs.NXmonitor(mode=scan_monitor_mode,
                                  preset=scan_acq_time)
        self.nxentry.insert(nxmonitor)
        monitor_data = self.ddfieldsDict[self.sanitizeName(
            env['monitor'])]  # to be linked later on

        # instrument group
        nxinstrument = nxs.NXinstrument()
        self.nxentry.insert(nxinstrument)

        # monochromator  group
        nxmonochromator = nxs.NXmonochromator()
        nxinstrument.insert(nxmonochromator)
        energy_data = self.ddfieldsDict[self.sanitizeName(
            env['monochromator'])]  # to be linked later on

        # incoming_beam  group
        nxincoming_beam = nxs.NXdetector(name='incoming_beam')
        nxinstrument.insert(nxincoming_beam)
        incbeam_data = self.ddfieldsDict[self.sanitizeName(
            env['incbeam'])]  # to be linked later on

        # absorbed_beam  group
        nxabsorbed_beam = nxs.NXdetector(name='absorbed_beam')
        nxinstrument.insert(nxabsorbed_beam)
        absbeam_data = self.ddfieldsDict[self.sanitizeName(
            env['absbeam'])]  # to be linked later on
        absbeam_data.attrs['signal'] = '1'
        absbeam_data.attrs['axes'] = 'energy'

        # source group
        nxsource = nxs.NXsource()
        nxinstrument.insert(nxsource)
        nxinstrument['source']['name'] = env.get(
            'SourceInfo', {}).get('name', 'Unknown')
        nxinstrument['source']['type'] = env.get(
            'SourceInfo', {}).get('type', 'Unknown')
        nxinstrument['source']['probe'] = env.get(
            'SourceInfo', {}).get('x-ray', 'Unknown')

        # data group
        nxdata = nxs.NXdata()
        self.nxentry.insert(nxdata)

        #@todo create the PreScanSnapshot
        # self._createPreScanSnapshot(env)

        # write everything to file
        self.nxentry.write()

        #@todo: do this with the PyTree api instead(how to do named links with the PyTree API????)
        self._nxln(monitor_data, nxmonitor, name='data')
        self._nxln(incbeam_data, nxincoming_beam, name='data')
        self._nxln(absbeam_data, nxabsorbed_beam, name='data')
        self._nxln(energy_data, nxmonochromator, name='energy')
        self._nxln(energy_data, nxdata, name='energy')
        self._nxln(absbeam_data, nxdata, name='absorbed_beam')

        self.nxentry.nxfile.flush()

    def _writeRecord(self, record):
        # most used variables in the loop
        fd, debug, warning = self.nxentry.nxfile, self.debug, self.warning
        nparray, npshape = numpy.array, numpy.shape
        rec_data, rec_nb = record.data, record.recordno

        for dd in self.datadesc:
            if dd.name in record.data:
                data = rec_data[dd.name]
                field = self.ddfieldsDict[dd.label]

                if data is None:
                    data = numpy.zeros(dd.shape, dtype=dd.dtype)
                if not hasattr(data, 'shape'):
                    data = nparray([data], dtype=dd.dtype)
                elif dd.dtype != data.dtype.name:
                    debug('%s casted to %s (was %s)', dd.label, dd.dtype,
                          data.dtype.name)
                    data = data.astype(dd.dtype)

                slab_offset = [rec_nb] + [0] * len(dd.shape)
                shape = [1] + list(npshape(data))
                try:
                    field.put(data, slab_offset, shape)
                    field.write()
                except:
                    warning("Could not write <%s> with shape %s", data, shape)
                    raise
            else:
                debug("missing data for label '%s'", dd.label)
        self.nxentry.nxfile.flush()

    def _endRecordList(self, recordlist):
        env = self.currentlist.getEnviron()
        self.nxentry.insert(nxs.NXfield(
            name='end_time', value=env['endtime'].isoformat()))
        # self._populateInstrumentInfo()
        # self._createNXData()
        self.nxentry.write()
        self.nxentry.nxfile.flush()
        self.debug("Finishing recording %d on file %s:",
                   env['serialno'], self.filename)
        return


#=========================================================================
# BEGIN: THIS BLOCK SHOULD BE REMOVED IF NEXUS ACCEPTS THE PATCH TO NXfield
#=========================================================================
try:
    from nxs import NXfield  # needs Nexus v>=4.3
    from nxs import napi, NeXusError

    class NXfield_comp(NXfield):

        # NOTE: THE CONSTRUCTOR IS OPTIONAL. IF NOT IMPLEMENTED, WE CAN STILL
        # USE THE nxslab_dims PROPERTY
        def __init__(self, value=None, name='field', dtype=None, shape=(), group=None,
                     attrs={}, nxslab_dims=None, **attr):
            NXfield.__init__(self, value=value, name=name, dtype=dtype, shape=shape, group=group,
                             attrs=attrs, **attr)
            self._slab_dims = nxslab_dims

        def write(self):
            """
            Write the NXfield, including attributes, to the NeXus file.
            """
            if self.nxfile:
                if self.nxfile.mode == napi.ACC_READ:
                    raise NeXusError("NeXus file is readonly")
                if not self.infile:
                    shape = self.shape
                    if shape == ():
                        shape = (1,)
                    with self.nxgroup as path:
                        if self.nxslab_dims is not None:
                            # compress
                            path.compmakedata(self.nxname, self.dtype, shape, 'lzw',
                                              self.nxslab_dims)
                        else:
                            # Don't use compression
                            path.makedata(self.nxname, self.dtype, shape)
                    self._infile = True
                if not self.saved:
                    with self as path:
                        path._writeattrs(self.attrs)
                        value = self.nxdata
                        if value is not None:
                            path.putdata(value)
                    self._saved = True
            else:
                raise IOError("Data is not attached to a file")

        def _getnxslabdims(self):
            try:
                return self._nxslab_dims
            except:
                slab_dims = None
            # even if slab_dims have not been set, check if the dataset is
            # large
            shape = self.shape or (1,)
            if numpy.prod(shape) > 10000:
                slab_dims = numpy.ones(len(shape), 'i')
                slab_dims[-1] = min(shape[-1], 100000)
            return slab_dims

        def _setnxslabdims(self, slab_dims):
            self._nxslab_dims = slab_dims

        nxslab_dims = property(_getnxslabdims, _setnxslabdims,
                               doc="Slab (a.k.a. chunk) dimensions for compression")
except:
    pass  # NXxas_FileRecorder won't be usable


#==============================================================================
# END: THE ABOVE BLOCK SHOULD BE REMOVED IF NEXUS ACCEPTS THE PATCH TO NXfield
#==============================================================================



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import threading
import math
import copy
import numpy
import traceback

from sardana import State
from sardana.sardanaevent import EventGenerator, EventType
from sardana.pool.pooldefs import SynchParam, SynchDomain
from taurus.core.util.log import Logger


def strictly_increasing(l):
    """Check whether list l has strictly increasing values"""
    return all(x < y for x, y in zip(l, l[1:]))


def strictly_decreasing(l):
    """Check whether list l has strictly deacreasing values"""
    return all(x > y for x, y in zip(l, l[1:]))


class FunctionGenerator(EventGenerator, Logger):
    """Generator of active and passive events describing a rectangular
    function.

    .. note::
        The FunctionGenerator class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    MAX_NAP_TIME = 0.1

    def __init__(self, name="FunctionGenerator"):
        EventGenerator.__init__(self)
        Logger.__init__(self, name)
        self._name = name
        self._initial_domain = None
        self._active_domain = None
        self._position_event = threading.Event()
        self._position = None
        self._initial_domain_in_use = None
        self._active_domain_in_use = None
        self._active_events = list()
        self._passive_events = list()
        self._started = False
        self._stopped = False
        self._running = False
        self._start_time = None
        self._direction = None
        self._condition = None
        self._id = None
        self._start_fired = False

    def get_name(self):
        return self._name

    name = property(get_name)

    def set_initial_domain(self, domain):
        self._initial_domain = domain

    def get_initial_domain(self):
        return self._initial_domain

    initial_domain = property(get_initial_domain, set_initial_domain)

    def set_active_domain(self, domain):
        self._active_domain = domain

    def get_active_domain(self):
        return self._active_domain

    active_domain = property(get_active_domain, set_active_domain)

    def set_initial_domain_in_use(self, domain):
        self._initial_domain_in_use = domain

    def get_initial_domain_in_use(self):
        return self._initial_domain_in_use

    initial_domain_in_use = property(get_initial_domain_in_use,
                                     set_initial_domain_in_use)

    def set_active_domain_in_use(self, domain):
        self._active_domain_in_use = domain

    def get_active_domain_in_use(self):
        return self._active_domain_in_use

    active_domain_in_use = property(get_active_domain_in_use,
                                    set_active_domain_in_use)

    def add_active_event(self, event):
        self._active_events.append(event)

    def set_active_events(self, events):
        self._active_events = events

    def get_active_events(self):
        return self._active_events

    active_events = property(get_active_events, set_active_events)

    def add_passive_event(self, event):
        self._passive_events.append(event)

    def set_passive_events(self, events):
        self._passive_events = events

    def get_passive_events(self):
        return self._passive_events

    passive_events = property(get_passive_events, set_passive_events)

    def set_direction(self, direction):
        self._direction = direction
        if direction == 1:
            self._condition = numpy.greater_equal
        elif direction == -1:
            self._condition = numpy.less_equal
        else:
            raise ValueError("direction can be -1 or 1 (negative or positive)")

    def get_direction(self):
        return self._direction

    direction = property(get_direction, set_direction)

    def event_received(self, *args, **kwargs):
        _, _, v = args
        if v.error:
            exc_info = v.exc_info
            self.error("Synchronization base attribute in error")
            msg = "Details: " + "".join(traceback.format_exception(*exc_info))
            self.debug(msg)
            return
        self._position = v.value
        self._position_event.set()

    def start(self):
        self._start_time = time.time()
        self._stopped = False
        self._started = True
        self._position = None
        self._start_fired = False
        self._position_event.clear()
        self._id = 0
        self.fire_event(EventType("state"), State.Moving)

    def stop(self):
        self._stopped = True

    def is_started(self):
        return self._started

    def is_stopped(self):
        return self._stopped

    def is_running(self):
        return self._running

    def run(self):
        self._running = True
        try:
            while len(self.active_events) > 0 and not self.is_stopped():
                self.wait_active()
                self.fire_active()
                self.wait_passive()
                self.fire_passive()
                self._id += 1
        finally:
            self._started = False
            self._running = False
            self._stopped = False
            self.fire_event(EventType("state"), State.On)

    def sleep(self, period):
        if period <= 0:
            return
        necessary_naps = int(math.ceil(period / self.MAX_NAP_TIME))
        if necessary_naps == 0:  # avoid zero ZeroDivisionError
            nap = 0
        else:
            nap = period / necessary_naps
        for _ in range(necessary_naps):
            if self.is_stopped():
                break
            time.sleep(nap)

    def fire_start(self):
        self.fire_event(EventType("start"), self._id)
        self._start_fired = True
        if self._id > 0:
            msg = "start was fired with {0} delay".format(self._id)
            self.warning(msg)

    def wait_active(self):
        candidate = self.active_events[0]
        if self.initial_domain_in_use == SynchDomain.Time:
            now = time.time()
            candidate += self._start_time
            self.sleep(candidate - now)
        else:
            while True:
                if self.is_stopped():
                    break
                if self._position_event.isSet():
                    self._position_event.clear()
                    now = self._position
                    if self._condition(now, candidate):
                        break
                else:
                    self._position_event.wait(self.MAX_NAP_TIME)

    def fire_active(self):
        # check if some events needs to be skipped
        i = 0
        while i < len(self.active_events) - 1:
            candidate = self.active_events[i + 1]
            if self.initial_domain_in_use is SynchDomain.Time:
                candidate += self._start_time
                now = time.time()
            elif self.initial_domain_in_use is SynchDomain.Position:
                now = self._position
            if self._condition(now, candidate):
                i += 1
            else:
                break
        self._id += i
        if not self._start_fired:
            self.fire_start()
        self.fire_event(EventType("active"), self._id)
        self.active_events = self.active_events[i + 1:]
        self.passive_events = self.passive_events[i:]

    def wait_passive(self):
        if self.active_domain_in_use == SynchDomain.Time:
            now = time.time()
            candidate = self._start_time + self.passive_events[0]
            self.sleep(candidate - now)
        else:
            while True:
                if self._position_event.isSet():
                    self._position_event.clear()
                    if self._condition(self._position, self.passive_events[0]):
                        break
                else:
                    self._position_event.wait(self.MAX_NAP_TIME)
                    if self.is_stopped():
                        break

    def fire_passive(self):
        self.fire_event(EventType("passive"), self._id)
        self.set_passive_events(self.passive_events[1:])
        if len(self.passive_events) == 0:
            self.fire_end()

    def fire_end(self):
        self.fire_event(EventType("end"), self._id)

    def set_configuration(self, configuration):
        # make a copy since we may inject the initial time
        configuration = copy.deepcopy(configuration)
        active_events = []
        passive_events = []
        self._direction = None
        # create short variables for commodity
        Time = SynchDomain.Time
        Position = SynchDomain.Position
        Initial = SynchParam.Initial
        Delay = SynchParam.Delay
        Active = SynchParam.Active
        Total = SynchParam.Total
        Repeats = SynchParam.Repeats

        for i, group in enumerate(configuration):
            # inject delay as initial time - generation will be
            # relative to the start time
            initial_param = group.get(Initial)
            if initial_param is None:
                initial_param = dict()
            if Time not in initial_param:
                delay_param = group.get(Delay)
                if Time in delay_param:
                    initial_param[Time] = delay_param[Time]
                group[Initial] = initial_param
            # determine active domain in use
            msg = "no initial value in group %d" % i
            if self.initial_domain in initial_param:
                self.initial_domain_in_use = self.initial_domain
            elif Position in initial_param:
                self.initial_domain_in_use = Position
            elif Time in initial_param:
                self.initial_domain_in_use = Time
            else:
                raise ValueError(msg)
            # determine passive domain in use
            active_param = group.get(Active)
            msg = "no active value in group %d" % i
            if self.active_domain is None:
                if Time in active_param:
                    self.active_domain_in_use = Time
                elif Position in active_param:
                    self.active_domain_in_use = Position
                else:
                    raise ValueError(msg)
            elif self.active_domain in active_param:
                self.active_domain_in_use = self.active_domain
            else:
                raise ValueError(msg)
            # create short variables for commodity
            initial_domain_in_use = self.initial_domain_in_use
            active_domain_in_use = self.active_domain_in_use
            repeats = group.get(Repeats, 1)
            active = active_param[active_domain_in_use]
            initial_in_initial_domain = initial_param[initial_domain_in_use]
            initial_in_active_domain = initial_param[active_domain_in_use]
            active_event_in_initial_domain = initial_in_initial_domain
            active_event_in_active_domain = initial_in_active_domain
            if repeats > 1:
                total_param = group[Total]
                total_in_initial_domain = total_param[initial_domain_in_use]
                total_in_active_domain = total_param[active_domain_in_use]
                for _ in range(repeats):
                    passive_event = active_event_in_active_domain + active
                    active_events.append(active_event_in_initial_domain)
                    passive_events.append(passive_event)
                    active_event_in_initial_domain += total_in_initial_domain
                    active_event_in_active_domain += total_in_active_domain
            else:
                active_events.append(active_event_in_initial_domain)
                passive_event = active_event_in_active_domain + active
                passive_events.append(passive_event)

        # determine direction
        if self.direction is None:
            if strictly_increasing(active_events):
                self.direction = 1
            elif strictly_decreasing(active_events):
                self.direction = -1
            else:
                msg = "active values indicate contradictory directions"
                raise ValueError(msg)

        self.active_events = active_events
        self.passive_events = passive_events



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["PseudoCounter", "PseudoCounterClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import Except, READ, SCALAR, SPECTRUM, IMAGE, DevDouble, \
    DevVarStringArray, DevVarDoubleArray, DevState, AttrQuality, DevFailed

from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.tango.core.util import to_tango_type_format, exception_str, \
    throw_sardana_exception
from sardana.tango.pool.PoolDevice import PoolExpChannelDevice, \
    PoolExpChannelDeviceClass


class PseudoCounter(PoolExpChannelDevice):

    def __init__(self, dclass, name):
        PoolExpChannelDevice.__init__(self, dclass, name)
        self._first_read_cache = False

    def init(self, name):
        PoolExpChannelDevice.init(self, name)

    def _is_allowed(self, req_type):
        return PoolExpChannelDevice._is_allowed(self, req_type)

    def get_pseudo_counter(self):
        return self.element

    def set_pseudo_counter(self, pseudo_counter):
        self.element = pseudo_counter

    pseudo_counter = property(get_pseudo_counter, set_pseudo_counter)

    @DebugIt()
    def delete_device(self):
        PoolExpChannelDevice.delete_device(self)
        pseudo_counter = self.pseudo_counter
        if pseudo_counter is not None:
            pseudo_counter.remove_listener(self.on_pseudo_counter_changed)

    @DebugIt()
    def init_device(self):
        PoolExpChannelDevice.init_device(self)

        self.Elements = list(map(int, self.Elements))
        pseudo_counter = self.pseudo_counter
        if pseudo_counter is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.pseudo_counter = pseudo_counter = \
                self.pool.create_element(type="PseudoCounter", name=name,
                                         full_name=full_name, id=self.Id, axis=self.Axis,
                                         ctrl_id=self.Ctrl_id, user_elements=self.Elements)
            if self.instrument is not None:
                pseudo_counter.set_instrument(self.instrument)
        pseudo_counter.add_listener(self.on_pseudo_counter_changed)

        self.set_state(DevState.ON)

    def on_pseudo_counter_changed(self, event_source, event_type,
                                  event_value):
        try:
            self._on_pseudo_counter_changed(event_source, event_type,
                                            event_value)
        except:
            msg = 'Error occurred "on_pseudo_counter_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.pseudo_counter.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_pseudo_counter_changed(self, event_source, event_type,
                                   event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name.lower()
        attr_name = name

        try:
            attr = self.get_attribute_by_name(attr_name)
        except DevFailed:
            return

        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        value, w_value, error = None, None, None

        if name == "state":
            value = self.calculate_tango_state(event_value)
        elif name == "status":
            value = self.calculate_tango_status(event_value)
        elif name == "valuebuffer":
            value = self._encode_value_chunk(event_value)
            self._first_read_cache = True
        elif name == "value":
            if isinstance(event_value, SardanaAttribute):
                # first obtain the value - during this process it may
                # enter into the error state, either when updating the elements
                # values or when calculating the pseudo value
                value = event_value.value
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                timestamp = event_value.timestamp
            else:
                value = event_value

            state = self.pseudo_counter.get_state(propagate=0)
            if state == State.Moving:
                quality = AttrQuality.ATTR_CHANGING
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                else:
                    value = event_value.value
                timestamp = event_value.timestamp

        self.set_attribute(attr, value=value, w_value=w_value,
                           timestamp=timestamp, quality=quality,
                           priority=priority, error=error, synch=False)

    def always_executed_hook(self):
        #state = to_tango_state(self.pseudo_counter.get_state(cache=False))
        pass

    def read_attr_hardware(self, data):
        pass

    def get_dynamic_attributes(self):
        cache_built = hasattr(self, "_dynamic_attributes_cache")

        std_attrs, dyn_attrs = \
            PoolExpChannelDevice.get_dynamic_attributes(self)

        if not cache_built:
            # For value attribute, listen to what the controller says for data
            # type (between long and float) and data format (scalar, spectrum
            # or image)
            value = std_attrs.get('value')
            if value is not None:
                _, data_info, attr_info = value
                ttype, tformat = to_tango_type_format(attr_info.dtype,
                                                      attr_info.dformat)
                data_info[0][0] = ttype
                data_info[0][1] = tformat
                if tformat == SPECTRUM:
                    shape = attr_info.maxdimsize
                    data_info[0].append(shape[0])
                elif tformat == IMAGE:
                    shape = attr_info.maxdimsize
                    data_info[0].append(shape[0])
                    data_info[0].append(shape[1])
        return std_attrs, dyn_attrs

    def initialize_dynamic_attributes(self):
        attrs = PoolExpChannelDevice.initialize_dynamic_attributes(self)

        detect_evts = "value",
        non_detect_evts = "valuebuffer",

        for attr_name in detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, True)
        for attr_name in non_detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, False)

    def read_Value(self, attr):
        pseudo_counter = self.pseudo_counter
        use_cache = pseudo_counter.is_in_operation() and not self.Force_HW_Read
        if not use_cache and self._first_read_cache:
            use_cache = True
            self._first_read_cache = False
        value_attr = pseudo_counter.get_value(cache=use_cache, propagate=0)
        # first obtain the value - during this process it may
        # enter into the error state, either when updating the elements
        # values or when calculating the pseudo value
        value = value_attr.value
        if value_attr.error:
            Except.throw_python_exception(*value_attr.exc_info)
        quality = None
        state = pseudo_counter.get_state(cache=use_cache, propagate=0)
        if state == State.Moving:
            quality = AttrQuality.ATTR_CHANGING
        timestamp = value_attr.timestamp
        self.set_attribute(attr, value=value, quality=quality,
                           priority=0, timestamp=timestamp)

    is_Value_allowed = _is_allowed

    def CalcPseudo(self, physical_values):
        """Returns the pseudo counter value for the given physical counters"""
        if not len(physical_values):
            physical_values = None
        result = self.pseudo_counter.calc(physical_values=physical_values)
        if result.error:
            throw_sardana_exception(result)
        return result.value

    def CalcAllPseudo(self, physical_values):
        """Returns the pseudo counter values for the given physical counters"""
        if not len(physical_values):
            physical_values = None
        result = self.pseudo_counter.calc(physical_values=physical_values)
        if result.error:
            throw_sardana_exception(result)
        return result.value


class PseudoCounterClass(PoolExpChannelDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
        "Elements":    [DevVarStringArray, "elements used by the pseudo", []],
    }
    device_property_list.update(PoolExpChannelDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'CalcPseudo': [[DevVarDoubleArray, "physical values"], [DevDouble, "pseudo counter"]],
        'CalcAllPseudo': [[DevVarDoubleArray, "physical positions"], [DevVarDoubleArray, "pseudo counter values"]],
    }
    cmd_list.update(PoolExpChannelDeviceClass.cmd_list)

    #    Attribute definitions
    standard_attr_list = {
        'Value': [[DevDouble, SCALAR, READ]],
    }
    standard_attr_list.update(PoolExpChannelDeviceClass.standard_attr_list)

    def _get_class_properties(self):
        ret = PoolExpChannelDeviceClass._get_class_properties(self)
        ret['Description'] = "Pseudo counter device class"
        ret['InheritedFrom'].insert(0, 'PoolExpChannelDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################


import uuid
import numpy

from taurus import Device
from unittest import TestCase
from taurus.test.base import insertTest
from sardana.sardanautils import is_number, is_non_str_seq, is_pure_str
from sardana.taurus.core.tango.sardana.pool import registerExtensions
from sardana.tango.pool.test.base_sartest import SarTestTestCase


def is_numerical(obj):
    if is_number(obj):
        return True
    if is_non_str_seq(obj) or isinstance(obj, numpy.ndarray):
        if is_number(obj[0]):
            return True
        elif is_non_str_seq(obj[0]) or isinstance(obj, numpy.ndarray):
            if is_number(obj[0][0]):
                return True
    return False


@insertTest(helper_name="count", test_method_doc="count with PC",
            elements=["_test_ct_1_1", "_test_ct_1_2", "_test_pc_1_1"])
@insertTest(helper_name="count", test_method_doc="count with Tango attribute",
            elements=["_test_ct_1_1", "_test_mt_1_1/position"])
@insertTest(helper_name="count", test_method_doc="count with 2D",
            elements=["_test_ct_1_1", "_test_2d_1_1"])
@insertTest(helper_name="count", test_method_doc="count with 1D",
            elements=["_test_ct_1_1", "_test_1d_1_1"])
@insertTest(helper_name="count", test_method_doc="count with 0D",
            elements=["_test_ct_1_1", "_test_0d_1_1"])
@insertTest(helper_name="count", test_method_doc="count with CT",
            elements=["_test_ct_1_1"])
class TestMeasurementGroup(SarTestTestCase, TestCase):

    def setUp(self):
        SarTestTestCase.setUp(self)
        registerExtensions()

    def count(self, elements):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            _, values = mg.count(.1)
            for channel_name, value in values.items():
                msg = "Value (%s) for %s is not numerical" % \
                      (value, channel_name)
                self.assertTrue(is_numerical(value), msg)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def tearDown(self):
        SarTestTestCase.tearDown(self)


class TestMeasurementGroupValueRef(SarTestTestCase, TestCase):

    def setUp(self):
        SarTestTestCase.setUp(self)
        registerExtensions()

    def test_value_ref_enabled(self):
        mg_name = str(uuid.uuid1())
        channel_name = "_test_2d_1_1"
        argin = [mg_name, channel_name]
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            channel = Device(channel_name)
            mg.setValueRefEnabled(True, channel_name)
            _, values = mg.count(.1)
            for channel_name, value in values.items():
                msg = "ValueRef (%s) for %s is not string" %\
                      (value, channel_name)
                self.assertTrue(is_pure_str(value), msg)
        finally:
            channel.cleanUp()
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_value_ref_disabled(self):
        mg_name = str(uuid.uuid1())
        channel_name = "_test_2d_1_1"
        argin = [mg_name, channel_name]
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            channel = Device(channel_name)
            mg.setValueRefEnabled(False, channel_name)
            _, values = mg.count(.1)
            for channel_name, value in values.items():
                msg = "Value (%s) for %s is not numerical" %\
                      (value, channel_name)
                self.assertTrue(is_numerical(value), msg)
        finally:
            channel.cleanUp()
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def tearDown(self):
        SarTestTestCase.tearDown(self)


class TestMotor(SarTestTestCase, TestCase):

    def setUp(self):
        SarTestTestCase.setUp(self)
        registerExtensions()

    def test_move(self):
        mot = Device("_test_mt_1_1")
        _, values = mot.move(1)

    def tearDown(self):
        SarTestTestCase.tearDown(self)
